# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import numpy as np

import SimpleITK as sitk

from platipy.imaging.label.utils import get_com

from platipy.imaging.utils.crop import crop_to_roi, label_to_roi

from platipy.imaging.utils.geometry import vector_angle

from platipy.imaging.utils.valve import generate_valve_using_cylinder


def extract(
    template_img,
    angles,
    radii,
    angle_min,
    angle_max,
    loc_x,
    loc_y,
    cw=False,
    radius_min=0,
    min_area_mm2=25,
):
    """
    Utility function to extract relevant voxels from a mask based on polar coordinates
    """
    # Get template array
    template_arr = sitk.GetArrayViewFromImage(template_img)
    # Get the segment array
    segment_arr = np.zeros_like(template_arr)

    # Define the condition list
    radius_min = radius_min

    if cw:
        in_segment_condition = (angles <= angle_min) | (angles >= angle_max)
        in_segment_condition &= radii >= radius_min
    else:
        in_segment_condition = (angles <= angle_max) & (angles >= angle_min)
        in_segment_condition &= radii >= radius_min

    # Extract matching voxels
    segment_arr[loc_y[in_segment_condition], loc_x[in_segment_condition]] = 1

    # Convert to image
    segment_img = sitk.GetImageFromArray(segment_arr)
    segment_img.CopyInformation(template_img)

    # Make sure area exceeds lower bound
    area = segment_arr.sum() * np.prod(segment_img.GetSpacing())
    if area < min_area_mm2:
        segment_img *= 0

    return sitk.Cast(segment_img, template_img.GetPixelID())


def generate_left_ventricle_segments(
    contours,
    label_left_ventricle="Ventricle_L",
    label_left_atrium="Atrium_L",
    label_right_ventricle="Ventricle_R",
    label_heart="Heart",
    myocardium_thickness_mm=10,
    hole_fill_mm=3,
    optimiser_tol_degrees=1,
    optimiser_max_iter=10,
    min_area_mm2=50,
    verbose=False,
):
    """
    Generates the 17 segments of the left vetricle

    This functions works as follows:
        1.  Heart volume is rotated to align the long axis to the z Cartesian (physical) space.
            Usually means it aligns with the axial axis (for normal simulation CT)
        2.  An optimiser adjusts the orientation to refine this alignment to the vector defined by
            MV COM - LV apex axis (long axis)
        3.  Left ventricle is divided into thirds along the long axis
        4.  Myocardium is defined as the outer 10mm
        5.  Geometric operations are used to define the segments
        6.  Everything is rotated back to the normal orientation
        7.  Some post-processing *magic*

    Args:
        contours (dict): A dictionary containing strings (label names) as keys and SimpleITK.Image
            (masks) as values. Must contain at least the LV, RV, MV, and whole heart.
        label_left_ventricle (str, optional): The name for the left ventricle mask (contour).
            Defaults to Ventricle_L.
        label_left_atrium (str, optional): The name for the left atrium mask (contour). Defaults to
            Atrium_L.
        label_right_ventricle (str, optional): The name for the right ventricle mask (contour).
            Defaults to Ventricle_R.
        label_heart (str, optional): The name for the heart mask (contour). Defaults to Heart.
        myocardium_thickness_mm (float, optional): Moycardial thickness, in millimetres.
            Defaults to 10.
        hole_fill_mm (float, optional): Holes smaller than this get filled in. Defaults to 3.
        optimiser_tol_degrees (float, optional): Optimiser tolerance (change in angle per iter).
            Defaults to 1, which typically requires 3-4 iterations.
        optimiser_max_iter (int, optional): Maximum optimiser iterations. Defaults to 10
        verbose (bool, optional): Print of information for debugging. Defaults to False.

    Returns:
        dict : The left ventricle segment dictionary, with labels (int) as keys and the binary
        label defining the segment (SimpleITK.Image) as values.
    """

    if verbose:
        print("Beginning LV segmentation algorithm.")

    # Initial set up
    label_mitral_valve = "MITRALVALVE"

    label_list = [label_left_ventricle, label_left_atrium, label_right_ventricle, label_heart]
    working_contours = copy.deepcopy({s: contours[s] for s in label_list})

    label_list.append(label_mitral_valve)

    output_contours = {}
    overall_transform_list = []

    # we add an automatically generated MV contour
    working_contours[label_mitral_valve] = generate_valve_using_cylinder(
        working_contours[label_left_atrium],
        working_contours[label_left_ventricle],
        radius_mm=15,
        height_mm=10,
    )

    # Some conversions
    erode_img = [
        int(myocardium_thickness_mm / i)
        for i in working_contours[label_left_ventricle].GetSpacing()
    ]
    hole_fill_img = [int(hole_fill_mm / i) for i in working_contours[label_heart].GetSpacing()]

    """
    Module 1 - Preparation
    Crop the images
    Rotate to the cardiac axis
    """
    # Crop to the smallest volume possible to make it FAST
    cb_size, cb_index = label_to_roi(
        working_contours[label_heart] > 0,
        expansion_mm=(30, 30, 60),  # Better to make it a bit bigger to be safe
    )

    for label in label_list:
        working_contours[label] = crop_to_roi(working_contours[label], cb_size, cb_index)

    if verbose:
        print("Module 1: Cropping and initial alignment.")
        vol_before = np.prod(contours[label_heart].GetSize())
        vol_after = np.prod(working_contours[label_heart].GetSize())
        print(f"  Images cropped. Volume reduction: {vol_before/vol_after:.3f}")

    # Initially we should reorient based on the cardiac axis
    label_orient = (
        working_contours[label_left_ventricle] + working_contours[label_left_atrium]
    ) > 0

    lsf = sitk.LabelShapeStatisticsImageFilter()  # this will be used throughout
    lsf.Execute(label_orient)
    cardiac_axis = np.array(lsf.GetPrincipalAxes(1)[:3])  # First principal axis approx. long axis

    # The principal axis isn't guaranteed to point from base to apex
    # If is points apex to base, we have to invert it
    # So check that here
    if cardiac_axis[2] < 0:
        cardiac_axis = -1 * cardiac_axis

    rotation_angle = vector_angle(cardiac_axis[::-1], (0, 0, 1))
    rotation_axis = np.cross(cardiac_axis[::-1], (0, 0, 1))
    rotation_centre = get_com(label_orient, real_coords=True)

    if verbose:
        print("  Alignment computed.")
        print("    Cardiac axis:    ", cardiac_axis)
        print("    Rotation axis:   ", rotation_axis)
        print("    Rotation angle:  ", rotation_angle)
        print("    Rotation centre: ", rotation_centre)

    rotation_transform = sitk.VersorRigid3DTransform()
    rotation_transform.SetCenter(rotation_centre)
    rotation_transform.SetRotation(rotation_axis, rotation_angle)

    overall_transform_list.append(rotation_transform)

    for label in label_list:
        working_contours[label] = sitk.Resample(
            working_contours[label],
            rotation_transform,
            sitk.sitkNearestNeighbor,
            0,
            working_contours[label].GetPixelID(),
        )

    """
    Module 2 - LV orientation alignment
    We use a very simple optimisation regime to enable robust computation of the LV apex
    We compute the vector from the MV COM to the LV apex
    This will be used for orientation (i.e. the long axis)
    """
    optimiser_tol_radians = optimiser_tol_degrees * np.pi / 180

    n = 0

    if verbose:
        print("Module 2: LV orientation alignment.")
        print("  Optimiser tolerance (degrees) =", optimiser_tol_degrees)
        print("  Beginning alignment process")

    while n < optimiser_max_iter and np.abs(rotation_angle) > optimiser_tol_radians:
        n += 1

        # Find the LV apex
        lv_locations = np.where(sitk.GetArrayViewFromImage(working_contours[label_left_ventricle]))
        lv_apex_z = lv_locations[0].min()
        lv_apex_y = lv_locations[1][lv_locations[0] == lv_apex_z].mean()
        lv_apex_x = lv_locations[2][lv_locations[0] == lv_apex_z].mean()
        lv_apex_loc = np.array([lv_apex_x, lv_apex_y, lv_apex_z])

        # Get the MV COM
        mv_com = np.array(get_com(working_contours[label_mitral_valve], real_coords=True))

        # Define the LV axis
        lv_apex_loc_img = np.array(
            working_contours[label_left_ventricle].TransformContinuousIndexToPhysicalPoint(
                lv_apex_loc.tolist()
            )
        )
        lv_axis = lv_apex_loc_img - mv_com

        # Compute the rotation parameters
        rotation_axis = np.cross(lv_axis, (0, 0, 1))
        rotation_angle = vector_angle(lv_axis, (0, 0, 1))
        rotation_centre = 0.5 * (
            mv_com + lv_apex_loc_img
        )  # get_com(working_contours[label_left_ventricle], real_coords=True)

        rotation_transform = sitk.VersorRigid3DTransform()
        rotation_transform.SetCenter(rotation_centre)
        rotation_transform.SetRotation(rotation_axis, rotation_angle)

        overall_transform_list.append(rotation_transform)

        if verbose:
            print("    N:               ", n)
            print("    LV apex:         ", lv_apex_loc_img)
            print("    MV COM:          ", mv_com)
            print("    LV axis:         ", lv_axis)
            print("    Rotation axis:   ", rotation_axis)
            print("    Rotation centre: ", rotation_centre)
            print("    Rotation angle:  ", rotation_angle)

        for label in label_list:
            working_contours[label] = sitk.Resample(
                working_contours[label],
                rotation_transform,
                sitk.sitkNearestNeighbor,
                0,
                working_contours[label].GetPixelID(),
            )

    """
    Module 3 - Compute the myocardium for the whole LV volume

    Divide this volume into thirds (from MV COM -> LV apex)        
    """

    if verbose:
        print("Module 3: Myocardium generation.")

    # First, let's just extract the myocardium
    label_lv_inner = sitk.BinaryErode(working_contours[label_left_ventricle], erode_img)
    label_lv_myo = working_contours[label_left_ventricle] - label_lv_inner

    # Mask the myo to a dilation of the blood pool
    # This helps improve shape consistency
    label_lv_myo_mask = sitk.BinaryDilate(label_lv_inner, erode_img)
    label_lv_myo = sitk.Mask(label_lv_myo, label_lv_myo_mask)

    # Computing limits for division into thirds
    # [xstart, ystart, zstart, xsize, ysize, zsize]
    # For the limits, we will use the centre of mass of the MV to the LV apex
    # The inner limit is used to assign the top portion (basal) of the LV to the anterior segment
    lsf.Execute(label_lv_inner)
    _, _, inf_limit_lv, _, _, extent = lsf.GetRegion(1)

    com_mv, _, _ = get_com(working_contours[label_mitral_valve])

    extent = com_mv - inf_limit_lv
    dc = int(extent / 3)

    # Define limits (cut LV into thirds)
    apical_extent = inf_limit_lv + dc
    mid_extent = inf_limit_lv + 2 * dc
    basal_extent = com_mv  # more complete coverage

    if verbose:
        print("  Apex (long axis) slice:      ", inf_limit_lv)
        print("  Apical section extent slice: ", apical_extent)
        print("  Mid section extent slice:    ", mid_extent)
        print("  Basal section extent slice:  ", basal_extent)
        print("    DeltaCut (DC): ", dc)
        print("    Extent:        ", extent)

    # Segment 17
    label_lv_myo_apex = label_lv_myo * 1  # make a copy
    label_lv_myo_apex[:, :, inf_limit_lv:] = 0

    # The apical segment
    label_lv_myo_apical = label_lv_myo * 1  # make a copy
    label_lv_myo_apical[:, :, :inf_limit_lv] = 0
    label_lv_myo_apical[:, :, apical_extent:] = 0

    # The mid segment
    label_lv_myo_mid = label_lv_myo * 1  # make a copy
    label_lv_myo_mid[:, :, :apical_extent] = 0
    label_lv_myo_mid[:, :, mid_extent:] = 0

    # The basal segment
    label_lv_myo_basal = label_lv_myo * 1  # make a copy
    label_lv_myo_basal[:, :, :mid_extent] = 0
    label_lv_myo_basal[:, :, basal_extent:] = 0

    """
    Module 4 - Generate 17 segments

        1. Find the basal (anterior) insertion of the RV
            This defines theta_0
        2. Find the baseline angle for the apical section
            This defines thera_0_apical
        3. Iterate though each section (apical, mid, basal):
            a. Convert each myocardium label loc to polar coords
            b. Assign each label to the appropriate LV segments
    """

    if verbose:
        print("Module 4: Segment generation.")

    # We need the angle for the basal RV insertion
    # This is the most counter-clockwise RV location
    # First, retrieve the most basal 5 slices
    loc_rv_z, loc_rv_y, loc_rv_x = np.where(
        sitk.GetArrayViewFromImage(working_contours[label_right_ventricle])
    )
    loc_rv_z_basal = np.arange(mid_extent, mid_extent + 5)

    if verbose:
        print("  RV basal slices: ", loc_rv_z_basal)

    theta_rv_insertion = []
    for z in loc_rv_z_basal:
        # Now get all the x and y positions
        loc_rv_basal_x = loc_rv_x[np.where(np.in1d(loc_rv_z, z))]
        loc_rv_basal_y = loc_rv_y[np.where(np.in1d(loc_rv_z, z))]

        # Now define the LV COM on each slice
        lv_com = get_com(working_contours[label_left_ventricle][:, :, int(z)])
        lv_com_basal_x = lv_com[1]
        lv_com_basal_y = lv_com[0]

        # Compute the angle
        theta_rv = np.arctan2(lv_com_basal_y - loc_rv_basal_y, loc_rv_basal_x - lv_com_basal_x)
        theta_rv[theta_rv < 0] += 2 * np.pi
        theta_rv_insertion.append(theta_rv.min())

    theta_0 = np.median(theta_rv_insertion)

    if verbose:
        print("  RV insertion angle (basal section): ", theta_0)

    # We also need the angle in the apical section for accurate segmentation
    lv_com_apical_list = []
    rv_com_apical_list = []
    for n in range(inf_limit_lv, apical_extent):
        lv_com_apical_list.append(get_com(working_contours[label_left_ventricle][:, :, n]))
        rv_com_apical_list.append(get_com(working_contours[label_right_ventricle][:, :, n]))

    lv_com_apical = np.mean(lv_com_apical_list, axis=0)
    rv_com_apical = np.mean(rv_com_apical_list, axis=0)

    theta_0_apical = np.arctan2(
        lv_com_apical[0] - rv_com_apical[0], rv_com_apical[1] - lv_com_apical[1]
    )

    if verbose:
        print(" Apical LV-RV COM angle: ", theta_0_apical)

    for i in range(17):
        working_contours[i + 1] = 0 * working_contours[label_heart]

    working_contours[17] = label_lv_myo_apex

    if verbose:
        print("  Computing apical segments")
    # We are now going to compute the segments in cylindical sections
    # First up - apical slices
    for n in range(inf_limit_lv, apical_extent):
        label_lv_myo_slice = label_lv_myo[:, :, n]

        # We will need numpy arrays here
        arr_lv_myo_slice = sitk.GetArrayViewFromImage(label_lv_myo_slice)
        loc_y, loc_x = np.where(arr_lv_myo_slice)

        # Now the origin
        y_0, x_0 = get_com(label_lv_myo_slice)

        # Compute the angle(s)
        theta = -np.arctan2(loc_y - y_0, loc_x - x_0) - theta_0_apical
        # Convert to [0,2*np.pi]
        theta[theta < 0] += 2 * np.pi

        # Compute the radii
        radii = np.sqrt((loc_y - y_0) ** 2 + (loc_x - x_0) ** 2)

        # Now assign to different segments
        working_contours[13][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            5 * np.pi / 4,
            7 * np.pi / 4,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[14][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            1 * np.pi / 4,
            7 * np.pi / 4,
            loc_x,
            loc_y,
            cw=True,
            min_area_mm2=min_area_mm2,
        )
        working_contours[15][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            1 * np.pi / 4,
            3 * np.pi / 4,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[16][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            3 * np.pi / 4,
            5 * np.pi / 4,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )

    if verbose:
        print("  Computing mid segments")
    # Second up - mid slices
    for n in range(apical_extent, mid_extent):
        label_lv_myo_slice = label_lv_myo[:, :, n]

        # We will need numpy arrays here
        arr_lv_myo_slice = sitk.GetArrayViewFromImage(label_lv_myo_slice)
        loc_y, loc_x = np.where(arr_lv_myo_slice)

        # Now the origin
        y_0, x_0 = get_com(label_lv_myo_slice)

        # Compute the angle(s)
        theta = -np.arctan2(loc_y - y_0, loc_x - x_0) - theta_0
        # Convert to [0,2*np.pi]
        theta[theta < 0] += 2 * np.pi

        # Compute the radii
        radii = np.sqrt((loc_y - y_0) ** 2 + (loc_x - x_0) ** 2)

        # Now assign to different segments
        working_contours[8][:, :, n] = extract(
            label_lv_myo_slice, theta, radii, 0, np.pi / 3, loc_x, loc_y, min_area_mm2=min_area_mm2
        )
        working_contours[9][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            1 * np.pi / 3,
            2 * np.pi / 3,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[10][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            2 * np.pi / 3,
            3 * np.pi / 3,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[11][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            3 * np.pi / 3,
            4 * np.pi / 3,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[12][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            4 * np.pi / 3,
            5 * np.pi / 3,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )
        working_contours[7][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            5 * np.pi / 3,
            2 * np.pi,
            loc_x,
            loc_y,
            min_area_mm2=min_area_mm2,
        )

    if verbose:
        print("  Computing basal segments")
    # Third up - basal slices
    for n in range(mid_extent, basal_extent):
        label_lv_myo_slice = label_lv_myo[:, :, n]

        # We will need numpy arrays here
        arr_lv_myo_slice = sitk.GetArrayViewFromImage(label_lv_myo_slice)
        loc_y, loc_x = np.where(arr_lv_myo_slice)

        if arr_lv_myo_slice.sum() == 0:
            continue

        # Now the origin
        y_0, x_0 = get_com(label_lv_myo_slice)

        # Compute the angle(s)
        theta = -np.arctan2(loc_y - y_0, loc_x - x_0) - theta_0
        # Convert to [0,2*np.pi]
        theta[theta < 0] += 2 * np.pi

        # Compute the radii
        radii = np.sqrt((loc_y - y_0) ** 2 + (loc_x - x_0) ** 2)

        # Now assign to different segments
        working_contours[2][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            0,
            np.pi / 3,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )
        working_contours[3][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            1 * np.pi / 3,
            2 * np.pi / 3,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )
        working_contours[4][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            2 * np.pi / 3,
            3 * np.pi / 3,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )
        working_contours[5][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            3 * np.pi / 3,
            4 * np.pi / 3,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )
        working_contours[6][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            4 * np.pi / 3,
            5 * np.pi / 3,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )
        working_contours[1][:, :, n] = extract(
            label_lv_myo_slice,
            theta,
            radii,
            5 * np.pi / 3,
            2 * np.pi,
            loc_x,
            loc_y,
            radius_min=15,
            min_area_mm2=min_area_mm2,
        )

    """
    Module 5 - re-orientation into image space

    We perform the total inverse transformation, and paste the labels back into the image space
    """

    if verbose:
        print("  Module 5: Re-orientation.")

    # Compute the total transform
    overall_transform = sitk.CompositeTransform(overall_transform_list)
    inverse_transform = overall_transform.GetInverse()

    # Rotate back to the original reference space
    for segment in range(17):
        new_structure = sitk.Resample(
            working_contours[segment + 1],
            inverse_transform,
            sitk.sitkNearestNeighbor,
            0,
        )

        if hole_fill_mm > 0:
            new_structure = sitk.BinaryMorphologicalClosing(new_structure, hole_fill_img)

        new_structure = sitk.Paste(
            contours[label_heart] * 0,
            new_structure,
            new_structure.GetSize(),
            (0, 0, 0),
            cb_index,
        )

        output_contours[f"Ventricle_L_Segment{segment + 1}"] = new_structure

    if verbose:
        print("Complete!")

    return output_contours



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from collections.abc import Iterable
import random

import SimpleITK as sitk

from platipy.imaging.generation.dvf import (
    generate_field_shift,
    generate_field_expand,
)

from platipy.imaging.generation.mask import (
    get_bone_mask,
)

from platipy.imaging.registration.utils import apply_transform


def apply_augmentation(image, augmentation, masks=[]):

    if not isinstance(image, sitk.Image):
        raise AttributeError("image should be a SimpleITK.Image")

    if isinstance(augmentation, DeformableAugment):
        augmentation = [augmentation]

    if not isinstance(augmentation, Iterable):
        raise AttributeError(
            "augmentation must be a DeformableAugment or an iterable (such as list) of"
            "DeformableAugment's"
        )

    transforms = []
    dvf = None
    for aug in augmentation:

        if not isinstance(aug, DeformableAugment):
            raise AttributeError("Each augmentation must be of type DeformableAugment")

        tfm, field = aug.augment()
        transforms.append(tfm)

        if dvf is None:
            dvf = field
        else:
            dvf += field

    transform = sitk.CompositeTransform(transforms)
    del transforms

    image_deformed = apply_transform(
        image,
        transform=transform,
        default_value=int(sitk.GetArrayViewFromImage(image).min()),
        interpolator=sitk.sitkLinear,
    )

    masks_deformed = []
    for mask in masks:
        masks_deformed.append(
            apply_transform(
                mask, transform=transform, default_value=0, interpolator=sitk.sitkNearestNeighbor
            )
        )

    if masks:
        return image_deformed, masks_deformed, dvf

    return image_deformed, dvf


def generate_random_augmentation(ct_image, masks):

    random.shuffle(masks)
    # mask_count = len(masks)
    # masks = masks[: random.randint(2, 5)]

    # print(len(masks))
    augmentation_types = [
        {
            "class": ShiftAugment,
            "args": {"vector_shift": [(-10, 10), (10, 10), (-10, 10)], "gaussian_smooth": (3, 5)},
        },
        {
            "class": ContractAugment,
            "args": {
                "vector_contract": [(0, 10), (0, 10), (0, 10)],
                "gaussian_smooth": (3, 5),
                "bone_mask": True,
            },
        },
        {
            "class": ExpandAugment,
            "args": {
                "vector_expand": [(0, 10), (0, 10), (0, 10)],
                "gaussian_smooth": (3, 5),
                "bone_mask": True,
            },
        },
    ]

    augmentation = []
    for mask in masks:
        aug = random.choice(augmentation_types)

        aug_class = aug["class"]
        aug_args = {}
        for arg in aug["args"]:

            value = aug["args"][arg]
            if isinstance(value, list):

                # Randomly sample for each dim
                result = []
                for rng in value:
                    result.append(random.randint(rng[0], rng[1]))
                value = result
            elif isinstance(value, tuple):
                # Randomly sample a value in range
                value = random.randint(value[0], value[1])

            if arg == "bone_mask" and aug["args"][arg]:
                value = get_bone_mask(ct_image)

            aug_args[arg] = value
        augmentation.append(aug_class(mask, **aug_args))
    return augmentation


class DeformableAugment(ABC):
    @abstractmethod
    def augment(self):

        # return deformation
        pass


class ShiftAugment(DeformableAugment):
    def __init__(self, mask, vector_shift=(10, 10, 10), gaussian_smooth=5):

        self.mask = mask
        self.vector_shift = vector_shift
        self.gaussian_smooth = gaussian_smooth

    def augment(self):

        _, transform, dvf = generate_field_shift(
            self.mask,
            self.vector_shift,
            self.gaussian_smooth,
        )
        return transform, dvf


class ExpandAugment(DeformableAugment):
    def __init__(self, mask, vector_expand=(10, 10, 10), gaussian_smooth=5, bone_mask=False):

        self.mask = mask
        self.vector_expand = vector_expand
        self.gaussian_smooth = gaussian_smooth
        self.bone_mask = bone_mask

    def augment(self):

        _, transform, dvf = generate_field_expand(
            self.mask,
            bone_mask=self.bone_mask,
            expand=self.vector_expand,
            gaussian_smooth=self.gaussian_smooth,
        )

        return transform, dvf


class ContractAugment(DeformableAugment):
    def __init__(self, mask, vector_contract=(10, 10, 10), gaussian_smooth=5, bone_mask=False):

        self.mask = mask
        self.contract = [int(-x / s) for x, s in zip(vector_contract, mask.GetSpacing())]
        self.gaussian_smooth = gaussian_smooth
        self.bone_mask = bone_mask

    def augment(self):

        _, transform, dvf = generate_field_expand(
            self.mask,
            bone_mask=self.bone_mask,
            expand=self.contract,
            gaussian_smooth=self.gaussian_smooth,
        )
        return transform, dvf



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import SimpleITK as sitk


def detect_holes(img, lower_threshold=-10000, upper_threshold=-400):
    """
    Detect all (air) holes in given image. Default threshold values given are defined for CT.

    Args:
        img (sitk.Image): The image in which to detect holes.
        lower_threshold (int, optional): The lower threshold of intensities. Defaults to -10000.
        upper_threshold (int, optional): The upper threshold of intensities. Defaults to -400.

    Returns:
        label_image: image containing labels of all holes detected
        labels: a list of holes detected and some of their properties
    """

    # Threshold to get holes
    btif = sitk.BinaryThresholdImageFilter()
    btif.SetInsideValue(1)
    btif.SetOutsideValue(0)
    btif.SetLowerThreshold(lower_threshold)
    btif.SetUpperThreshold(upper_threshold)
    holes = btif.Execute(img)

    # Get connected components
    ccif = sitk.ConnectedComponentImageFilter()
    label_image = ccif.Execute(holes)

    labels = []

    lssif = sitk.LabelShapeStatisticsImageFilter()
    lssif.Execute(label_image)
    for region in range(1, ccif.GetObjectCount()):
        label = {
            "label": region,
            "phys_size": lssif.GetPhysicalSize(region),
            "elongation": lssif.GetElongation(region),
            "roundness": lssif.GetRoundness(region),
            "perimeter": lssif.GetPerimeter(region),
            "flatness": lssif.GetFlatness(region),
        }
        labels.append(label)

    # Sort by size
    labels = sorted(labels, key=lambda i: i["phys_size"], reverse=True)

    return label_image, labels


def get_external_mask(label_image, labels, kernel_radius=5):
    """Get a External mask given a label image and labels computed with detect_holes

    Args:
        label_image (sitk.Image): Label image from detect_holes
        labels (list): List of labels from detect_holes
        kernel_radius (int, optional): The width of the radius used by binary close. Defaults to 5.

    Returns:
        sitk.Image: External mask
    """

    bmcif = sitk.BinaryMorphologicalClosingImageFilter()
    bmcif.SetKernelType(sitk.sitkBall)
    bmcif.SetKernelRadius(kernel_radius)

    # Save the largest as the external contour
    external_mask = sitk.BinaryThreshold(label_image, labels[0]["label"], labels[0]["label"])
    external_mask = bmcif.Execute(external_mask)

    return external_mask


def get_lung_mask(label_image, labels, kernel_radius=2):
    """Get a Lung mask given a label image and labels computed with detect_holes

    Args:
        label_image (sitk.Image): Label image from detect_holes
        labels (list): List of labels from detect_holes
        kernel_radius (int, optional): The width of the radius used by binary close. Defaults to 2.

    Returns:
        sitk.Image: Lung mask
    """

    lung_idx = 1
    while labels[lung_idx]["flatness"] > 2:
        lung_idx += 1

        if lung_idx > len(labels):
            print("Flatness not satisfied!")
            return None

    bmcif = sitk.BinaryMorphologicalClosingImageFilter()
    bmcif.SetKernelType(sitk.sitkBall)
    bmcif.SetKernelRadius(kernel_radius)

    # Save the 2nd largest as the lung mask
    lung_mask = sitk.BinaryThreshold(
        label_image, labels[lung_idx]["label"], labels[lung_idx]["label"]
    )
    lung_mask = bmcif.Execute(lung_mask)

    return lung_mask


def fill_holes(img, label_image, external_mask, lung_mask, fill_value=50):
    """Returns the input image with all holes filled (except for the external and lung holes).

    Args:
        img (sitk.Image): The image to fill
        label_image (sitk.Image): Label image from detect_holes
        external_mask (sitk.Image): The external mask of the patient.
        lung_mask (sitk.Image): The lung mask of the patient.
        fill_value (int): The value to use to fill the holes. Defaults to 50.

    Returns:
        sitk.Image: The image with holes filled
    """

    img_array = sitk.GetArrayFromImage(img)

    bdif = sitk.BinaryDilateImageFilter()
    bdif.SetKernelType(sitk.sitkBall)
    bdif.SetKernelRadius(3)

    mask = sitk.BinaryThreshold(label_image, 1, int(img_array.max()))
    mask = sitk.Subtract(mask, external_mask)
    mask = sitk.Subtract(mask, lung_mask)
    mask = bdif.Execute(mask)

    mask_array = sitk.GetArrayFromImage(mask)
    img_array[mask_array == 1] = fill_value

    filled_img = sitk.GetImageFromArray(img_array)
    filled_img.CopyInformation(img)

    return filled_img



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
from multiprocessing import Process
import uuid

from sqlalchemy.exc import OperationalError
import click
import redis

from .models import db, APIKey


def create_db():
    """
    Create the Database used by the framework
    """

    db.create_all()


def add_api_key(name, is_admin):
    """
    Add a new API Key with the given name

    name: Name of the API Key
    """

    ak = APIKey(key=str(uuid.uuid4()), name=name, is_admin=is_admin)
    db.session.add(ak)
    db.session.commit()

    return ak.key


@click.group()
@click.pass_context
def cli(_):
    pass


@cli.command()
@click.pass_context
@click.option("--add", "-a", help="Name of api key to add")
@click.option("--list", "-l", is_flag=True, help="List all API keys")
@click.option(
    "--super",
    "-s",
    is_flag=True,
    help="API Key has super user priviledges (has access to other application's data)",
)
def key(ctx, add, list, super):

    try:
        keys = APIKey.query.all()

        if add in [k.name for k in keys]:
            click.echo("Error: An API Key with that name already exists.")
            return

        if add:
            api_key = add_api_key(add, super)

            click.echo("Key added: {0}".format(api_key))

        if list:
            for ak in APIKey.query.all():
                click.echo(ak)
    except OperationalError as oe:
        click.echo("An error occurred. Be sure to init the db with the 'initdb' command.")


@cli.command()
@click.pass_context
def initdb(ctx):

    create_db()

    click.echo("DB Created")


@cli.command()
@click.pass_context
@click.option("--ip", "-ip", help="IP address of service", required=True)
def ssl(ctx, ip):

    command = [
        "openssl",
        "req",
        "-new",
        "-x509",
        "-days",
        "365",
        "-nodes",
        "-subj",
        f"/CN={ip}",
        "-addext",
        f"subjectAltName = IP:{ip}",
        "-out",
        "service.crt",
        "-keyout",
        "service.key",
    ]

    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()

    click.echo(stdout)
    click.echo(stderr)

    click.echo("SSL Certificates Created. Service restart required.")


def run_celery():

    celery_command = [
        "celery",
        "--app=service:celery",
        "beat",
        "--loglevel=INFO",
    ]

    process = subprocess.Popen(celery_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print("Running celery worker & beat")
    stdout, stderr = process.communicate()

    click.echo(stdout)
    click.echo(stderr)


@cli.command()
@click.pass_context
@click.option("--port", "-p", help="Port on which to run service", default=8000)
def run(ctx, port):

    r = redis.client.Redis()
    try:
        r.ping()
    except redis.exceptions.ConnectionError:
        click.echo("redis is not running. Start with command: `redis-server --daemonize yes`")
        return

    process_celery = Process(target=run_celery)
    process_celery.start()

    service_command = [
        "gunicorn",
        "-b",
        f":{port}",
        "--timeout",
        "300",
        "--graceful-timeout",
        "60",
        "service:app",
    ]

    process = subprocess.Popen(
        f"gunicorn -b :{port} service:app",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    click.echo("Running Service (with gunicorn)")
    process.communicate()


if __name__ == "__main__":
    cli(obj={})



# Copyright 2020 CSIRO

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from platipy.dicom.io.nifti_to_rtstruct import convert_nifti


@click.command()
@click.option(
    "--dcm_file",
    "-d",
    required=True,
    help="Reference DICOM file from which header tags will be copied",
)
@click.option("--mask", "-m", multiple=True, required=True, help="Mask pairs with name,filename")
@click.option("--out_rt_filename", "-o", required=True, help="Name of RT struct output")
def click_command(dcm_file, mask, out_rt_filename):
    """
    Convert Nifti masks to Dicom RTStruct
    """

    convert_nifti(dcm_file, mask, out_rt_filename)


if __name__ == "__main__":
    click_command()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python

# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging
import json

import click

import SimpleITK as sitk

from platipy.imaging.projects.bronchus.run import (
    run_bronchus_segmentation,
    BRONCHUS_SETTINGS_DEFAULTS,
)

logger = logging.getLogger(__name__)

segmentation_algorithms = {}

try:
    from platipy.imaging.projects.cardiac.run import (
        run_hybrid_segmentation,
        HYBRID_SETTINGS_DEFAULTS,
    )

    segmentation_algorithms["cardiac"] = {
        "algorithm": run_hybrid_segmentation,
        "default_settings": HYBRID_SETTINGS_DEFAULTS,
    }
except ImportError:
    logger.warning(
        "Unable to provide cardiac segmentation. Be sure to install platipy with "
        "cardiac extras: pip install 'platipy[cardiac]'"
    )

segmentation_algorithms["bronchus"] = {
    "algorithm": run_bronchus_segmentation,
    "default_settings": BRONCHUS_SETTINGS_DEFAULTS,
}


@click.command()
@click.argument("algorithm", nargs=1, type=click.Choice(segmentation_algorithms.keys()))
@click.argument("input_path", nargs=1, type=click.Path(), required=False)
@click.option(
    "--config",
    "-c",
    required=False,
    type=click.Path(),
    help="Path to JSON file containing algorithm settings",
)
@click.option(
    "--default",
    "-d",
    is_flag=True,
    help="Print the default configuration for the selected algorithm",
)
@click.option(
    "--output",
    "-o",
    required=False,
    type=click.Path(),
    help="Path to directory for output",
)
def click_command(algorithm, input_path, config, default, output):
    """
    Run an auto-segmentation on an input image. Choose which algorithm to run and pass the path to
    the Nitfti input image OR a directory containing a DICOM series.

    Optionally, pass in a configuration file for the segmentation algorithm. Output the default
    configuration for an algorithm using the --default flag.
    """

    algorithm_config = segmentation_algorithms[algorithm]["default_settings"]

    if default:
        print(json.dumps(algorithm_config, indent=4))
        return

    # If we get to here but no input_path was set, we need to inform the user
    if not input_path:
        print("Supply the path the the Nifti Image OR DICOM series to process")
        return

    print(f"Running {algorithm} segmentation")

    if config:
        with open(config, "r") as file_obj:
            algorithm_config = json.load(file_obj)

    read_path = input_path
    if os.path.isdir(input_path):
        # If it's a directory then read as a DICOM series
        read_path = sitk.ImageSeriesReader().GetGDCMSeriesFileNames(input_path)

    image = sitk.ReadImage(read_path)

    # Run the algorithm
    results = segmentation_algorithms[algorithm]["algorithm"](image, algorithm_config)

    # If results is a tuple, the algorithm is returning other items as well as the semgmentation.
    # We are only interested in the segmentation here (the first element of the tuple)
    if isinstance(results, tuple):
        results = results[0]

    # Save the output to the output directory
    if not output:
        output = "."
    if not os.path.exists(output):
        os.makedirs(output, exist_ok=True)

    for result in results:
        sitk.WriteImage(results[result], os.path.join(output, f"{result}.nii.gz"))


if __name__ == "__main__":

    click_command()  # pylint: disable=no-value-for-parameter



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import shutil
import tempfile

import imageio

import numpy as np
import SimpleITK as sitk

import matplotlib

MATPLOTLIB_VERSION = matplotlib.__version__.split(".")
MATPLOTLIB_MAJOR = int(MATPLOTLIB_VERSION[0])
MATPLOTLIB_MINOR = int(MATPLOTLIB_VERSION[1])

if MATPLOTLIB_MAJOR >= 3 and MATPLOTLIB_MINOR >= 4:
    raise SystemError("Animation functionality requires a matplotlib version lower than 3.4.0")

import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.animation import FileMovieWriter


class FileWriter(FileMovieWriter):
    """Class to write image frames for animation"""

    supported_formats = ["png"]

    # pylint: disable=arguments-differ, attribute-defined-outside-init
    def setup(self, fig, dpi, frame_prefix):
        super().setup(fig, dpi, frame_prefix)
        self.fname_format_str = "%s%%d.%s"
        self.temp_prefix, self.frame_format = self.outfile.split(".")

    def grab_frame(self, **savefig_kwargs):
        with self._frame_sink() as myframesink:
            self.fig.savefig(myframesink, format="png", dpi=self.dpi, **savefig_kwargs)

    def finish(self):
        self._frame_sink().close()


def generate_animation_from_image_sequence(
    image_list,
    output_file="animation.gif",
    fps=10,
    contour_list=None,
    scalar_list=None,
    figure_size_in=6,
    image_cmap=matplotlib.colormaps.get_cmap("Greys_r"),
    contour_cmap=matplotlib.colormaps.get_cmap("jet"),
    scalar_cmap=matplotlib.colormaps.get_cmap("magma"),
    image_window=[-1000, 800],
    scalar_min=None,
    scalar_max=None,
    scalar_alpha=0.5,
    image_origin="lower",
    contour_kwargs={},
):
    """Generates an animation from a list of images, with optional scalar overlay and contours.

    Args:
        image_list (list (SimpleITK.Image)): A list of SimpleITK (2D) images.
        output_file (str, optional): The name of the output file. Defaults to "animation.gif".
        fps (int, optional): Frames per second. Defaults to 10.
        contour_list (list (SimpleITK.Image), optional): A list of SimpleITK (2D) images
            (overlay as scalar field). Defaults to False.
        scalar_list (list (SimpleITK.Image), optional): A list of SimpleITK (2D) images
            (overlay as contours). Defaults to False.
        figure_size_in (int, optional): Size of the figure. Defaults to 6.
        image_cmap (matplotlib.colors.ListedColormap, optional): Colormap to use for the image.
            Defaults to matplotlib.colormaps.get_cmap("Greys_r").
        contour_cmap (matplotlib.colors.ListedColormap, optional): Colormap to use for contours.
            Defaults to matplotlib.colormaps.get_cmap("jet").
        scalar_cmap (matplotlib.colors.ListedColormap, optional): Colormap to use for scalar field.
            Defaults to matplotlib.colormaps.get_cmap("magma").
        image_window (list, optional): Image intensity window (mininmum, range).
            Defaults to [-1000, 800].
        scalar_min (bool, optional): Minimum scalar value to show. Defaults to False.
        scalar_max (bool, optional): Maximum scalar value to show. Defaults to False.
        scalar_alpha (float, optional): Alpha (transparency) for scalar field. Defaults to 0.5.
        image_origin (str, optional): Image origin. Defaults to "lower".

    Raises:
        RuntimeError: If ImageMagick isn't installed you cannot use this function!
        ValueError: The list of images must be of type SimpleITK.Image

    Returns:
        matplotlib.animation: The animation.
    """

    if not isinstance(image_list[0], sitk.Image):
        raise ValueError("Each image must be a SimplITK image (sitk.Image).")

    # Get the image information
    x_size, y_size = image_list[0].GetSize()
    x_spacing, y_spacing = image_list[1].GetSpacing()

    asp = y_spacing / x_spacing

    # Define the figure
    figure_size = (figure_size_in, figure_size_in * (asp * y_size) / (1.0 * x_size))
    fig, ax = plt.subplots(1, 1, figsize=(figure_size))

    # Display the first image
    # This will be updated
    display_image = ax.imshow(
        sitk.GetArrayFromImage(image_list[0]),
        aspect=asp,
        interpolation=None,
        origin=image_origin,
        cmap=image_cmap,
        clim=(image_window[0], image_window[0] + image_window[1]),
    )

    # We now deal with the contours
    # These can be given as a list of sitk.Image objects or a list of dicts {"name":sitk.Image}
    if contour_list:
        if isinstance(contour_list[0], sitk.Image):
            plot_dict = {"_": contour_list[0]}
            contour_labels = False
        else:
            plot_dict = contour_list[0]
            contour_labels = True

        if isinstance(contour_cmap, dict):
            color_list = [contour_cmap[i] for i in plot_dict.keys()]
        else:
            color_map = contour_cmap(np.linspace(0, 1, len(plot_dict)))
            color_list = [color_map[i] for i in range(len(plot_dict.values()))]

        for ctr_key, color in zip(plot_dict.keys(), color_list):
            display_contours = ax.contour(
                sitk.GetArrayFromImage(plot_dict[ctr_key]),
                colors=[color],
                levels=[1],
                **contour_kwargs,
            )

            display_contours.collections[0].set_label(ctr_key)

        if contour_labels:
            approx_scaling = figure_size_in / (len(plot_dict.keys()))
            ax.legend(
                loc="upper left",
                bbox_to_anchor=(0.025, 0.975),
                fontsize=min([10, 16 * approx_scaling]),
            )

    if scalar_list:
        if not scalar_min:
            scalar_min = np.min([sitk.GetArrayFromImage(i) for i in scalar_list])
        if not scalar_max:
            scalar_max = np.max([sitk.GetArrayFromImage(i) for i in scalar_list])

        display_scalar = ax.imshow(
            np.ma.masked_outside(sitk.GetArrayFromImage(scalar_list[0]), scalar_min, scalar_max),
            aspect=asp,
            interpolation=None,
            origin=image_origin,
            cmap=scalar_cmap,
            clim=(scalar_min, scalar_max),
            alpha=scalar_alpha,
            vmin=scalar_min,
            vmax=scalar_max,
        )

    ax.axis("off")
    fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

    # The animate function does (you guessed it) the animation
    def animate(i):
        # Update the imaging data
        nda = sitk.GetArrayFromImage(image_list[i])
        display_image.set_data(nda)

        # TO DO - add in code for scalar overlay
        if contour_list:
            try:
                ax.collections = []
            except ValueError:
                pass

            if not isinstance(contour_list[i], dict):
                plot_dict = {"_": contour_list[i]}
            else:
                plot_dict = contour_list[i]

            if isinstance(contour_cmap, dict):
                color_list = [contour_cmap[i] for i in plot_dict.keys()]
            else:
                color_map = contour_cmap(np.linspace(0, 1, len(plot_dict)))
                color_list = [color_map[i] for i in range(len(plot_dict.values()))]

            for contour, color in zip(plot_dict.values(), color_list):
                ax.contour(
                    sitk.GetArrayFromImage(contour),
                    colors=[color],
                    levels=[1],
                    **contour_kwargs,
                )

        if scalar_list:
            nda = sitk.GetArrayFromImage(scalar_list[i])
            display_scalar.set_data(np.ma.masked_outside(nda, scalar_min, scalar_max))

        return (display_image,)

    # create animation using the animate() function with no repeat
    animation_result = animation.FuncAnimation(
        fig,
        animate,
        frames=np.arange(0, len(image_list), 1),
        interval=1,
        blit=True,
        repeat=False,
    )

    # Save animation
    tmp_path = tempfile.mkdtemp()
    animation_result.save(f"{tmp_path}/tmp.format", writer=FileWriter())

    # Save the GIF
    images = []
    for filename in sorted(
        pathlib.Path(tmp_path).glob("tmp*.png"), key=lambda n: int(n.name[3:-4])
    ):
        try:
            images.append(imageio.imread(filename))
        except RuntimeError:
            # Skip frames which are corrupt
            pass

    imageio.mimsave(
        output_file,
        images,
        duration=1 / fps,
    )

    # Clean up
    shutil.rmtree(tmp_path)

    return animation_result



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

import SimpleITK as sitk

import pandas as pd

import matplotlib
import matplotlib.lines as mlines
import matplotlib as plt
import matplotlib.colors as mcolors


from platipy.imaging.label.comparison import (
    compute_metric_dsc,
    compute_metric_hd,
    compute_metric_masd,
    compute_volume,
)

from platipy.imaging.label.utils import get_com
from platipy.imaging import ImageVisualiser


def contour_comparison(
    img,
    contour_dict_a,
    contour_dict_b,
    contour_label_a="Set A",
    contour_label_b="Set B",
    s_select=None,
    structure_for_com=None,
    structure_for_limits=None,
    title="",
    subtitle="",
    subsubtitle="",
    contour_cmap=matplotlib.colormaps.get_cmap("rainbow"),
    structure_name_dict=None,
    img_vis_kw=None,
):
    """Generates a custom figure for comparing two sets of contours (delineations) on an image.

    Includes a formatted table of contouring (similarity) metrics.

    Args:
        img (SimpleITK.Image): The image to visualise.
        contour_dict_a (dict): A dict object with contour names (str) as keys and SimpleITK.Image
            as values.
        contour_dict_b (dict): A dict object with contour names (str) as keys and SimpleITK.Image
            as values.
        contour_label_a (str, optional): Name to put in legend. Defaults to "Set A".
        contour_label_b (str, optional): Name to put in legend. Defaults to "Set B".
        s_select (list, optional): A list of contours to display. Defaults to all keys common to
            both contour sets.
        structure_for_com (str, optional): Name of the contour to set slice location.
            Defaults to the largest structure by volume.
        structure_for_limits (str, optional): Name of the contour to set view limits.
         Defaults to None, and entire image is displayed.
        title (str, optional): Title of the plot, set to "" to remove. Defaults to "TITLE TEXT".
        subtitle (str, optional): Subitle of the plot, set to "" to remove.
            Defaults to "SUBTITLE TEXT".
        subsubtitle (str, optional): Subsubtitle of the plot, set to "" to remove.
            Defaults to "SUBSUBTITLE TEXT".
        contour_cmap (plt.cm.colormap, optional): Contour colormap.
            Defaults to matplotlib.colormaps.get_cmap("rainbow").
        structure_name_dict (dict, optional): A "translation" dictionary used to overwrite the
            names of contours in the metric table. Defaults to using the names in the contour_dict.
        img_vis_kw (dict, optional): Passed to the ImageVisualiser class. Defaults to {}.

    Returns:
        matplotlib.figure: The figure, can be saved as usual (fig.savefig(...)).
    """

    # If no contour names are seleted we just use those which both contour_dicts have
    if s_select is None:
        s_select = [i for i in contour_dict_a.keys() if i in contour_dict_b.keys()]

    if img_vis_kw is None:
        img_vis_kw = {}

    if "cut" not in img_vis_kw:
        cut = None

        # first check if the user has specified a structure
        if structure_for_com is None:
            # If there is no option for the COM structure we use the largest
            # first calculate the "volume" of each structure (actually the sum of voxels)
            s_vol = [sitk.GetArrayFromImage(contour_dict_a[s]).sum() for s in s_select]

            if sum(s_vol) == 0:
                # if all structures are zero, try the second contour set
                s_vol = [sitk.GetArrayFromImage(contour_dict_b[s]).sum() for s in s_select]

                if sum(s_vol) == 0:
                    # if all of these structures are also zero, we don't have any contours!
                    cut = None

                else:
                    # otherwise, get the COM of the largest structure (in contour_set_b)
                    cut = get_com(contour_dict_b[s_select[np.argmax(s_vol)]])
            else:
                # otherwise, get the COM of the largest structure (in contour_set_a)
                cut = get_com(contour_dict_a[s_select[np.argmax(s_vol)]])

        else:
            # the user has specified a structure
            # first, check the structure isn't empty!
            if sitk.GetArrayFromImage(contour_dict_a[structure_for_com]).sum() != 0:
                cut = get_com(contour_dict_a[structure_for_com])
            # if it is, try the same structure in contour_set_b
            elif sitk.GetArrayFromImage(contour_dict_b[structure_for_com]).sum() != 0:
                cut = get_com(contour_dict_b[structure_for_com])

        img_vis_kw["cut"] = cut

    # Colormap options
    if isinstance(contour_cmap, (mcolors.ListedColormap, mcolors.LinearSegmentedColormap)):
        colors_a = {
            s + "a": c
            for s, c in zip(
                s_select,
                matplotlib.colormaps.get_cmap(contour_cmap)(np.linspace(0, 1, len(s_select))),
            )
        }

        colors_b = {
            s + "b": c
            for s, c in zip(
                s_select,
                matplotlib.colormaps.get_cmap(contour_cmap)(np.linspace(0, 1, len(s_select))),
            )
        }

    elif isinstance(contour_cmap, dict):
        colors_a = {s + "a": contour_cmap[s] for s in s_select}
        colors_b = {s + "b": contour_cmap[s] for s in s_select}

    # Visualise!
    vis = ImageVisualiser(img, **img_vis_kw)

    # Add contour set A
    vis.add_contour(
        {s + "a": contour_dict_a[s] for s in s_select},
        show_legend=False,
        color=colors_a,
    )

    # Add contour set B
    vis.add_contour(
        {s + "b": contour_dict_b[s] for s in s_select},
        show_legend=False,
        color=colors_b,
        linestyle="dashed",
    )

    # Optionally, set limits
    if structure_for_limits is not None:
        vis.set_limits_from_label(contour_dict_a[structure_for_limits], expansion=20)

    # Create the figure
    fig = vis.show()

    # Choose the blank axis (top right)
    ax = fig.axes[1]

    # Set row names is structure_name_dict is given
    if structure_name_dict is not None:
        rows = [structure_name_dict[i] for i in s_select]
    else:
        rows = s_select

    # Names
    suffix_a = "A"
    suffix_b = "B"
    if contour_label_a != "Set A":  # the default
        suffix_a = contour_label_a
    if contour_label_b != "Set A":  # the default
        suffix_b = contour_label_b

    # Compute some metrics
    df_metrics = pd.DataFrame(
        columns=[
            "STRUCTURE",
            "DSC",
            "MDA_mm",
            "HD_mm",
            f"VOL_{suffix_a}_cm3",
            f"VOL_{suffix_b}_cm3",
        ]
    )
    columns = ("DSC", "MDA\n[mm]", "HD\n[mm]", "Vol.\nRatio")

    cell_text = []

    for s, row in zip(s_select, rows):
        dsc = compute_metric_dsc(contour_dict_a[s], contour_dict_b[s])
        mda = compute_metric_masd(contour_dict_a[s], contour_dict_b[s])
        hd = compute_metric_hd(contour_dict_a[s], contour_dict_b[s])
        vol_a = compute_volume(contour_dict_a[s])
        vol_b = compute_volume(contour_dict_b[s])

        cell_text.append(
            [
                f"{dsc:.2f}",
                f"{mda:.2f}",
                f"{hd:.2f}",
                f"{vol_b/vol_a:.2f}",
            ]
        )

        # compute metrics and add to dataframe
        row = pd.DataFrame(
            [
                {
                    "STRUCTURE": s,
                    "DSC": dsc,
                    "MDA_mm": mda,
                    "HD_mm": hd,
                    f"VOL_{suffix_a}_cm3": vol_a,
                    f"VOL_{suffix_b}_cm3": vol_b,
                }
            ]
        )
        df_metrics = pd.concat([df_metrics, row])

    # If there are no labels we can make the table bigger
    if title == "" and subsubtitle == "" and subsubtitle == "":
        v_extent = 0.88
    else:
        v_extent = 0.7

    v_extent = min([v_extent, 0.1 * len(list(contour_dict_a.keys()))])

    # Create the table
    table = ax.table(
        cellText=cell_text,
        rowLabels=rows,
        rowColours=list(colors_a.values()),
        colLabels=columns,
        fontsize=10,
        bbox=[0.25, 0.1, 0.73, v_extent],
    )

    # Some nice formatting
    for cell in table.get_celld():
        table[cell].set_text_props(va="center")
        table[cell].set_edgecolor("w")

        if cell[0] == 0:  # header
            table[cell].set_text_props(weight="bold", color="w")
            table[cell].set_facecolor("k")

    # Geometry fixes
    for row in range(len(rows) + 1):
        table[row, 0].set_width(0.1)
        table[row, 1].set_width(0.1)
        table[row, 2].set_width(0.1)
        table[row, 3].set_width(0.1)
        if row > 0:
            table[row, -1].set_width(0)

    for col in range(len(columns)):
        table[0, col].set_height(0.075)

    table.auto_set_font_size(True)
    fs = table.get_celld()[1, 0].get_fontsize()

    # Insert text
    ax.text(
        0.95,
        0.98,
        title,
        color="navy",
        ha="right",
        va="top",
        size=fs + 4,
    )
    ax.text(0.95, 0.92, subtitle, color="darkgreen", ha="right", va="top", size=fs + 2)
    ax.text(0.95, 0.87, subsubtitle, color="k", ha="right", va="top", size=fs + 2)

    # Insert legend
    _solid = mlines.Line2D([], [], color="k", label=contour_label_a)
    _dashed = mlines.Line2D([], [], color="k", linestyle="dashed", label=contour_label_b)
    ax.legend(
        handles=[_solid, _dashed],
        bbox_to_anchor=(0.25, 0.02, 0.73, 0.1),
        ncol=2,
        mode="expand",
        borderaxespad=0.0,
        fontsize=fs,
        loc="lower left",
    )

    # Return the figure
    return fig, df_metrics



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import os
import json
from pprint import pformat

import requests

logger = logging.getLogger(__name__)


API_DICOM_LOCATION = "{0}/api/dicomlocation"
API_DATASET = "{0}/api/dataset"
API_DATASET_READY = "{0}/api/dataset/ready"
API_DATA_OBJECT = "{0}/api/dataobject"
API_TRIGGER = "{0}/api/trigger"
API_ALGORITHM = "{0}/api/algorithm"
API_DOWNLOAD_OBJECT = "{0}/api/dataobject/download"


class PlatiPyClient:
    """Client to help iteract with the framework implemented within PlatiPy."""

    def __init__(self, host, port, api_key, algorithm_name, verify=None):
        """Initialize an instance of the client. Will test to ensure that the host it reachable

        Arguments:
            host {str} -- IP or Host name of server running service
            port {int} -- Port number server is using to running service
            api_key {str} -- The API Key created for this app on the server
            algorithm_name {str} -- The name fo the algorithm to use within the service
        """

        protocol = "https"

        if verify is None:
            logger.warning("Running without SSL. Not Suitable for Production.")
            protocol = "http"
        else:
            if not os.path.exists(verify):
                raise FileNotFoundError("Verify Certificate file not found")

        self.verify = verify

        self.base_url = f"{protocol}://{host}:{port}"

        logger.info("Initializing client with URL: %s", self.base_url)

        self.api_key = api_key
        self.algorithm_name = algorithm_name

        res = requests.get(
            API_ALGORITHM.format(self.base_url),
            headers={"API_KEY": self.api_key},
            verify=self.verify,
        )
        logger.debug(res.status_code)

    def get_dicom_location(self, name):
        """Gets the dicom location with the name supplied

        Arguments:
            name {str} -- The name of the dicom location

        Returns:
            dict -- The dicom location dictionary object, or None if it doesn't exist
        """

        res = requests.get(
            API_DICOM_LOCATION.format(self.base_url),
            headers={"API_KEY": self.api_key},
            verify=self.verify,
        )
        logger.debug(res.status_code)

        for location in res.json():
            if location["name"] == name:
                logger.debug(pformat(location))
                return location

        return None

    def add_dicom_location(self, name, host, port, ae_title=None):
        """Adds a new DICOM location.

        Arguments:
            name {str} -- The name representing this DICOM location
            host {str} -- The IP or host name of the DICOM location
            port {int} -- The port this DICOM location communicates on

        Keyword Arguments:
            ae_title {str} -- The AETitle of the DICOM location (default: {None})

        Returns:
            dict -- The DICOM location dictionary object, or None if something went wr
        """

        location = self.get_dicom_location(name)
        if location:
            logger.info("Location with name '%s' already exists", name)
            return location

        params = {"name": name, "host": host, "port": port}

        if ae_title:
            params["ae_title"] = ae_title

        res = requests.post(
            API_DICOM_LOCATION.format(self.base_url),
            headers={"API_KEY": self.api_key},
            data=params,
            verify=self.verify,
        )
        logger.debug(res.status_code)

        if res.status_code >= 200 and res.status_code < 300:
            logger.info("Added Location")
            location = res.json()
            logger.debug(pformat(location))
            return location

        return None

    def get_dataset(self, dataset):
        """Fetches a dataset from the server

        Arguments:
            dataset {dict/int} -- The dictionary object representing the dataset or the id of the
                                  dataset

        Returns:
            dict -- The dictionary object representing the dataset
        """

        params = {"dataset": dataset}

        if isinstance(dataset, dict):
            params["dataset"] = dataset["id"]

        res = requests.get(
            "{0}/{1}".format(API_DATASET.format(self.base_url), params["dataset"]),
            headers={"API_KEY": self.api_key},
            verify=self.verify,
        )
        logger.debug(res.status_code)

        if res.status_code == 200:
            dataset = res.json()
            logger.debug(pformat(dataset))
            return dataset

        return None

    def get_dataset_ready(self, dataset):
        """Checks to see if the dataset is ready to run on the server

        Arguments:
            dataset {dict/int} -- The dictionary object representing the dataset or the id of the
                                  dataset

        Returns:
            dict -- The dictionary object representing the dataset
        """

        params = {"dataset": dataset}

        if isinstance(dataset, dict):
            params["dataset"] = dataset["id"]

        res = requests.get(
            "{0}/{1}".format(API_DATASET_READY.format(self.base_url), params["dataset"]),
            headers={"API_KEY": self.api_key},
            verify=self.verify,
        )
        logger.debug(res.status_code)

        if res.status_code == 200:
            result = res.json()
            logger.debug(pformat(result))
            return result["ready"]

        return None

    def add_dataset(self, from_dicom_location=None, to_dicom_location=None, timeout=None):
        """Adds and returns a new dataset

        Keyword Arguments:
            from_dicom_location {dict/int} -- The DICOM location to fetch objects from
                                              (default: {None})
            to_dicom_location {dict/int} -- The DICOM location to send resulting objects to
                                            (default: {None})
            timeout {int} -- How long the data within the dataset should be kept on the server
                             (default: {None})

        Returns:
            dict -- The resulting dictionary object representing the dataset or None if something
                    went wrong
        """

        params = {}

        if from_dicom_location:

            params["from_dicom_location"] = from_dicom_location

            if isinstance(from_dicom_location, dict):
                params["from_dicom_location"] = from_dicom_location["id"]

        if to_dicom_location:

            params["to_dicom_location"] = to_dicom_location

            if isinstance(to_dicom_location, dict):
                params["to_dicom_location"] = to_dicom_location["id"]

        if timeout:
            params["timeout"] = timeout

        res = requests.post(
            API_DATASET.format(self.base_url),
            headers={"API_KEY": self.api_key},
            data=params,
            verify=self.verify,
        )
        logger.debug(res.status_code)

        if res.status_code >= 200 and res.status_code < 300:
            logger.info("Added Dataset")
            dataset = res.json()
            logger.debug(pformat(dataset))
            return dataset

        return None

    def add_data_object(
        self,
        dataset,
        series_uid=None,
        parent=None,
        meta_data=None,
        dicom_retrieve=None,
        file_path=None,
    ):
        """Adds and returns an input data object to a dataset specified

        Arguments:
            dataset {dict/int} -- The dataset to which to add the data object

        Keyword Arguments:
            series_uid {str} -- The SeriesInstanceUID of the DICOM object to fetch (required for
                                DICOM objects) (default: {None})
            parent {dict/int} -- The parent object for this dataobject (default: {None})
            meta_data {dict} -- The Meta Data to attach to this data object (default: {None})
            dicom_retrieve {str} -- How the DICOM object should be fetched ['GET', 'MOVE' or
                                    'SEND'] (required for DICOM objects) (default: {None})
            file_path {str} -- Path to the file to add as the data object (Required for non-DICOM
                               objects) (default: {None})

        Returns:
            dict -- The resulting dictionary representing the data object or None if something
                    went wrong
        """

        data_object = None

        params = {"dataset": dataset}
        if isinstance(dataset, dict):
            params["dataset"] = dataset["id"]

        if parent:
            params["parent"] = parent
            if isinstance(parent, dict):
                params["parent"] = parent["id"]

        if meta_data:
            params["meta_data"] = json.dumps(meta_data)

        if series_uid or dicom_retrieve:

            if not series_uid or dicom_retrieve:
                logger.error("For Dicom, both series_uid and dicom_retrieve must be set")
                return None

            params["type"] = "DICOM"
            params["seriesUID"] = series_uid
            params["dicom_retrieve"] = dicom_retrieve

            res = requests.post(
                API_DATA_OBJECT.format(self.base_url),
                headers={"API_KEY": self.api_key},
                data=params,
                verify=self.verify,
            )
            logger.debug(res.status_code)

            if res.status_code >= 200:
                data_object = res.json()
        else:

            if not file_path:
                logger.error("For a file, provide the file_path")
                return None

            params["type"] = "FILE"
            params["file_name"] = os.path.basename(file_path)

            with open(file_path, "rb") as file_handle:

                res = requests.post(
                    API_DATA_OBJECT.format(self.base_url),
                    headers={"API_KEY": self.api_key},
                    data=params,
                    files={"file_data": file_handle},
                    verify=self.verify,
                )
                logger.debug(res.status_code)

                if res.status_code >= 200:
                    data_object = res.json()

        logger.debug(pformat(data_object))
        return data_object

    def get_default_settings(self):
        """Gets the default settings for the algorithm

        Returns:
            dict -- Dictionary object with the default settings, None if something went wrong
        """

        algorithm = None
        res = requests.get(
            API_ALGORITHM.format(self.base_url),
            headers={"API_KEY": self.api_key},
            verify=self.verify,
        )
        logger.debug(res.status_code)
        if res.status_code == 200:
            for algorithm in res.json():
                if self.algorithm_name in algorithm["name"]:
                    logger.debug(pformat(algorithm))
                    if "default_settings" in algorithm:
                        return algorithm["default_settings"]

                    # If no default settings show an error
                    logger.error("No default_settings provided by algorithm")
        return None

    def run_algorithm(self, dataset, config=None):
        """Runs the algorithm on the dataset specified

        Arguments:
            dataset {dict/int} -- The dataset on which to run the algorithm

        Yields:
            dict -- Status of the run
        """

        params = {"dataset": dataset, "algorithm": self.algorithm_name}

        if isinstance(dataset, dict):
            params["dataset"] = dataset["id"]

        if config:

            # Check that the keys of the config passed in are the exact same as the default keys
            default_settings = self.get_default_settings()

            if not set(default_settings.keys()) == set(config.keys()):
                logger.error("Config keys must be exactly those from the default_settings")
                return

            params["config"] = json.dumps(config)

        res = requests.post(
            API_TRIGGER.format(self.base_url),
            headers={"API_KEY": self.api_key},
            data=params,
            verify=self.verify,
        )
        logger.debug(res.status_code)

        if res.status_code == 200:
            # Poll the URL given to determine the progress of the task
            poll_url = "{0}{1}".format(self.base_url, res.json()["poll"])

            while True:
                res = requests.get(poll_url, headers={"API_KEY": self.api_key}, verify=self.verify)
                status = res.json()

                if (
                    not "state" in status
                    or status["state"] == "SUCCESS"
                    or status["state"] == "FAILURE"
                ):
                    break

                yield status

                time.sleep(1)
        else:
            logger.error(res.json())

        logger.info("Algorithm Processing Complete")

    def download_output_objects(self, dataset, output_path="."):
        """Downloads the output objects from the dataset specified to the output path

        Arguments:
            dataset {dict/int} -- The dataset from which the output objects should be downloaded

        Keyword Arguments:
            output_path {str} -- The directory in which the objects should be downloaded
                                 (default: {"."})
        """

        if not os.path.exists(output_path):
            logger.info("Creating directory")
            os.makedirs(output_path)

        dataset = self.get_dataset(dataset)
        if dataset:
            for data_obj in dataset["output_data_objects"]:
                url = API_DOWNLOAD_OBJECT.format(self.base_url)
                res = requests.get(
                    "{0}/{1}".format(url, data_obj["id"]),
                    headers={"API_KEY": self.api_key},
                    verify=self.verify,
                )
                logger.debug(res.status_code)
                filename = res.headers["Content-Disposition"].split("filename=")[1]

                output_file = os.path.join(output_path, filename)
                logger.info("Downloading to: %s", output_file)
                open(output_file, "wb").write(res.content)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

import SimpleITK as sitk
import numpy as np

from platipy.imaging.registration.utils import apply_transform

from platipy.imaging.registration.linear import (
    linear_registration,
)

from platipy.imaging.registration.deformable import (
    fast_symmetric_forces_demons_registration,
)

from platipy.imaging.label.fusion import (
    process_probability_image,
    compute_weight_map,
    combine_labels,
)

from platipy.imaging.utils.crop import label_to_roi, crop_to_roi

from platipy.imaging.label.utils import correct_volume_overlap

logger = logging.getLogger(__name__)

ATLAS_PATH = "/atlas"
if "ATLAS_PATH" in os.environ:
    ATLAS_PATH = os.environ["ATLAS_PATH"]

MUTLIATLAS_SETTINGS_DEFAULTS = {
    "atlas_settings": {
        "atlas_id_list": [
            "03",
        ],
        "atlas_structure_list": [
            "WHOLEHEART",
        ],
        "atlas_path": ATLAS_PATH,
        "atlas_image_format": "Case_{0}/Images/Case_{0}_CROP.nii.gz",
        "atlas_label_format": "Case_{0}/Structures/Case_{0}_{1}_CROP.nii.gz",
        "crop_atlas_to_structures": False,
        "crop_atlas_expansion_mm": (20, 20, 40),
    },
    "auto_crop_target_image_settings": {
        "expansion_mm": [20, 20, 40],
    },
    "linear_registration_settings": {
        "reg_method": "affine",
        "shrink_factors": [16, 8, 4],
        "smooth_sigmas": [0, 0, 0],
        "sampling_rate": 0.75,
        "default_value": None,
        "number_of_iterations": 50,
        "metric": "mean_squares",
        "optimiser": "gradient_descent_line_search",
        "verbose": False,
    },
    "deformable_registration_settings": {
        "isotropic_resample": True,
        "resolution_staging": [
            6,
            3,
            1.5,
        ],  # specify voxel size (mm) since isotropic_resample is set
        "iteration_staging": [150, 125, 100],
        "smoothing_sigmas": [
            0,
            0,
            0,
        ],
        "ncores": 8,
        "default_value": None,
        "verbose": False,
    },
    "label_fusion_settings": {
        "vote_type": "unweighted",
        "vote_params": None,
        "optimal_threshold": {},
    },
    "postprocessing_settings": {
        "run_postprocessing": True,
        "binaryfillhole_mm": 3,
        "structures_for_binaryfillhole": [],
        "structures_for_overlap_correction": [],
    },
}


def run_segmentation(img, settings=MUTLIATLAS_SETTINGS_DEFAULTS):
    """Runs the atlas-based segmentation algorithm

    Args:
        img (sitk.Image):
        settings (dict, optional): Dictionary containing settings for algorithm.
                                   Defaults to default_settings.

    Returns:
        dict: Dictionary containing output of segmentation
    """

    results = {}
    results_prob = {}

    """
    Initialisation - Read in atlases
    - image files
    - structure files

        Atlas structure:
        'ID': 'Original': 'CT Image'    : sitk.Image
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image
                'RIR'     : 'CT Image'    : sitk.Image
                            'Transform'   : transform parameter map
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image
                'DIR'     : 'CT Image'    : sitk.Image
                            'Transform'   : displacement field transform
                            'Weight Map'  : sitk.Image
                            'Struct A'    : sitk.Image
                            'Struct B'    : sitk.Image


    """

    logger.info("")
    # Settings
    atlas_path = settings["atlas_settings"]["atlas_path"]
    atlas_id_list = settings["atlas_settings"]["atlas_id_list"]
    atlas_structure_list = settings["atlas_settings"]["atlas_structure_list"]

    atlas_image_format = settings["atlas_settings"]["atlas_image_format"]
    atlas_label_format = settings["atlas_settings"]["atlas_label_format"]

    crop_atlas_to_structures = settings["atlas_settings"]["crop_atlas_to_structures"]
    crop_atlas_expansion_mm = settings["atlas_settings"]["crop_atlas_expansion_mm"]

    atlas_set = {}
    for atlas_id in atlas_id_list:
        atlas_set[atlas_id] = {}
        atlas_set[atlas_id]["Original"] = {}

        image = sitk.ReadImage(f"{atlas_path}/{atlas_image_format.format(atlas_id)}")

        structures = {
            struct: sitk.ReadImage(f"{atlas_path}/{atlas_label_format.format(atlas_id, struct)}")
            for struct in atlas_structure_list
        }

        if crop_atlas_to_structures:
            logger.info("Automatically cropping atlas: %s", atlas_id)

            original_volume = np.prod(image.GetSize())

            crop_box_size, crop_box_index = label_to_roi(
                structures.values(), expansion_mm=crop_atlas_expansion_mm
            )

            image = crop_to_roi(image, size=crop_box_size, index=crop_box_index)

            final_volume = np.prod(image.GetSize())

            logger.info("  > Volume reduced by factor %.2f", original_volume/final_volume)

            for struct in atlas_structure_list:
                structures[struct] = crop_to_roi(
                    structures[struct], size=crop_box_size, index=crop_box_index
                )

        atlas_set[atlas_id]["Original"]["CT Image"] = image

        for struct in atlas_structure_list:
            atlas_set[atlas_id]["Original"][struct] = structures[struct]

    """
    Step 1 - Automatic cropping
    If we have a guide structure:
        - use structure to crop target image

    Otherwise:
        - using a quick registration to register each atlas
        - expansion of the bounding box to ensure entire volume of interest is enclosed
        - target image is cropped
    """

    expansion_mm = settings["auto_crop_target_image_settings"]["expansion_mm"]

    quick_reg_settings = {
        "reg_method": "similarity",
        "shrink_factors": [8],
        "smooth_sigmas": [0],
        "sampling_rate": 0.75,
        "default_value": -1000,
        "number_of_iterations": 25,
        "final_interp": sitk.sitkLinear,
        "metric": "mean_squares",
        "optimiser": "gradient_descent_line_search",
    }

    registered_crop_images = []

    logger.info("Running initial Translation tranform to crop image volume")

    for atlas_id in atlas_id_list[: min([8, len(atlas_id_list)])]:

        logger.info("  > atlas %s", atlas_id)

        # Register the atlases
        atlas_set[atlas_id]["RIR"] = {}
        atlas_image = atlas_set[atlas_id]["Original"]["CT Image"]

        reg_image, _ = linear_registration(
            img,
            atlas_image,
            **quick_reg_settings,
        )

        registered_crop_images.append(sitk.Cast(reg_image, sitk.sitkFloat32))

        del reg_image

    combined_image = sum(registered_crop_images) / len(registered_crop_images) > -1000

    crop_box_size, crop_box_index = label_to_roi(combined_image, expansion_mm=expansion_mm)

    img_crop = crop_to_roi(img, crop_box_size, crop_box_index)

    logger.info("Calculated crop box:")
    logger.info("  > %s", crop_box_index)
    logger.info("  > %s", crop_box_size)
    logger.info("  > Vol reduction = %.2f", np.prod(img.GetSize())/np.prod(crop_box_size))

    """
    Step 2 - Rigid registration of target images
    - Individual atlas images are registered to the target
    - The transformation is used to propagate the labels onto the target
    """
    linear_registration_settings = settings["linear_registration_settings"]

    logger.info(
        "Running %s tranform to align atlas images", linear_registration_settings['reg_method']
    )

    for atlas_id in atlas_id_list:
        # Register the atlases

        logger.info("  > atlas %s", atlas_id)

        atlas_set[atlas_id]["RIR"] = {}

        target_reg_image = img_crop
        atlas_reg_image = atlas_set[atlas_id]["Original"]["CT Image"]

        _, initial_tfm = linear_registration(
            target_reg_image,
            atlas_reg_image,
            **linear_registration_settings,
        )

        # Save in the atlas dict
        atlas_set[atlas_id]["RIR"]["Transform"] = initial_tfm

        atlas_set[atlas_id]["RIR"]["CT Image"] = apply_transform(
            input_image=atlas_set[atlas_id]["Original"]["CT Image"],
            reference_image=img_crop,
            transform=initial_tfm,
            default_value=-1000,
            interpolator=sitk.sitkLinear,
        )

        # sitk.WriteImage(rigid_image, f"./RR_{atlas_id}.nii.gz")

        for struct in atlas_structure_list:
            input_struct = atlas_set[atlas_id]["Original"][struct]
            atlas_set[atlas_id]["RIR"][struct] = apply_transform(
                input_image=input_struct,
                reference_image=img_crop,
                transform=initial_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

        atlas_set[atlas_id]["Original"] = None

    """
    Step 3 - Deformable image registration
    - Using Fast Symmetric Diffeomorphic Demons
    """

    # Settings
    deformable_registration_settings = settings["deformable_registration_settings"]

    logger.info("Running DIR to refine atlas image registration")

    for atlas_id in atlas_id_list:

        logger.info("  > atlas %s", atlas_id)

        # Register the atlases
        atlas_set[atlas_id]["DIR"] = {}

        atlas_reg_image = atlas_set[atlas_id]["RIR"]["CT Image"]
        target_reg_image = img_crop

        _, dir_tfm, _ = fast_symmetric_forces_demons_registration(
            target_reg_image,
            atlas_reg_image,
            **deformable_registration_settings,
        )

        # Save in the atlas dict
        atlas_set[atlas_id]["DIR"]["Transform"] = dir_tfm

        atlas_set[atlas_id]["DIR"]["CT Image"] = apply_transform(
            input_image=atlas_set[atlas_id]["RIR"]["CT Image"],
            transform=dir_tfm,
            default_value=-1000,
            interpolator=sitk.sitkLinear,
        )

        for struct in atlas_structure_list:
            input_struct = atlas_set[atlas_id]["RIR"][struct]
            atlas_set[atlas_id]["DIR"][struct] = apply_transform(
                input_image=input_struct,
                transform=dir_tfm,
                default_value=0,
                interpolator=sitk.sitkNearestNeighbor,
            )

        atlas_set[atlas_id]["RIR"] = None

    """
    Step 4 - Label Fusion
    """
    # Compute weight maps
    vote_type = settings["label_fusion_settings"]["vote_type"]
    vote_params = settings["label_fusion_settings"]["vote_params"]

    # Compute weight maps
    for atlas_id in list(atlas_set.keys()):
        atlas_image = atlas_set[atlas_id]["DIR"]["CT Image"]
        weight_map = compute_weight_map(
            img_crop, atlas_image, vote_type=vote_type, vote_params=vote_params
        )
        atlas_set[atlas_id]["DIR"]["Weight Map"] = weight_map

    combined_label_dict = combine_labels(atlas_set, atlas_structure_list)

    """
    Step 6 - Paste the cropped structure into the original image space
    """
    logger.info("Generating binary segmentations.")
    template_img_binary = sitk.Cast((img * 0), sitk.sitkUInt8)
    template_img_prob = sitk.Cast((img * 0), sitk.sitkFloat64)

    for structure_name in atlas_structure_list:

        probability_map = combined_label_dict[structure_name]

        if structure_name in settings["label_fusion_settings"]["optimal_threshold"]:
            optimal_threshold = settings["label_fusion_settings"]["optimal_threshold"][
                structure_name
            ]
        else:
            optimal_threshold = 0.5

        binary_struct = process_probability_image(probability_map, optimal_threshold)

        # Un-crop binary structure
        paste_img_binary = sitk.Paste(
            template_img_binary,
            binary_struct,
            binary_struct.GetSize(),
            (0, 0, 0),
            crop_box_index,
        )
        results[structure_name] = paste_img_binary

        # Un-crop probability map
        paste_prob_img = sitk.Paste(
            template_img_prob,
            probability_map,
            probability_map.GetSize(),
            (0, 0, 0),
            crop_box_index,
        )
        results_prob[structure_name] = paste_prob_img

    """
    Step 8 - Post-processing
    """
    postprocessing_settings = settings["postprocessing_settings"]

    if postprocessing_settings["run_postprocessing"]:
        logger.info("Running post-processing.")

        # Remove any smaller components and perform morphological closing (hole filling)
        binaryfillhole_img = [
            int(postprocessing_settings["binaryfillhole_mm"] / sp) for sp in img.GetSpacing()
        ]

        for structure_name in postprocessing_settings["structures_for_binaryfillhole"]:

            if structure_name not in results:
                continue

            contour_s = results[structure_name]
            contour_s = sitk.RelabelComponent(sitk.ConnectedComponent(contour_s)) == 1
            contour_s = sitk.BinaryMorphologicalClosing(contour_s, binaryfillhole_img)
            results[structure_name] = contour_s

        if len(postprocessing_settings["structures_for_overlap_correction"]) >= 2:
            # Remove any overlaps
            input_overlap = {
                s: results[s] for s in postprocessing_settings["structures_for_overlap_correction"]
            }
            output_overlap = correct_volume_overlap(input_overlap)

            for s in postprocessing_settings["structures_for_overlap_correction"]:
                results[s] = output_overlap[s]

    logger.info("Done!")

    return results, results_prob



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1 import make_axes_locatable  # , AxesGrid, ImageGrid

import numpy as np
import SimpleITK as sitk

import logging
logger = logging.getLogger(__name__)


class View:
    """
    Class to display a single 2D view comprising images, contours, vectors, etc
    """

    def __init__(
        self,
        image,
        axis=None,
        contours=None,
        bounding_boxes=None,
        scalar_overlays=None,
        vector_overlays=None,
        crosshair_overlay=None
    ):

    if isinstance(image,)
    self.__arr_img = sitk.GetArrayFromImage
    self.__dimension = self.__arr_img.ndim

    def create_slice(
        self,
        axis_slice=None,
    ):
        """Set the slice (location) at which to cut the image/contours/overlays.
        This should be called first to put all the visuals onto an axis.

        Args:
            slice (int, optional): The slice location. Defaults to None.
        """

        ax = self.__axis
        aspect

        if arr_self.__arr_img.ndim == 2:
            logger.info("Cannot set a slice on a 2D image, ignoring.")
            
            axis_slicer = return_slice()

        if axis_slice is None:
            if arr_self.__arr_img.ndim == 2:
                axis_slice = 

        # First: display the image on the axis

        image = self.__image

        ax.imshow(
            arr_slice,
            aspect=aspect
        )

    def update_slice(
        self,
        slice=None
    ):



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['platipy',
 'platipy.backend',
 'platipy.backend.sample',
 'platipy.cli',
 'platipy.dicom',
 'platipy.dicom.communication',
 'platipy.dicom.download',
 'platipy.dicom.io',
 'platipy.dicom.tests',
 'platipy.imaging',
 'platipy.imaging.dose',
 'platipy.imaging.generation',
 'platipy.imaging.label',
 'platipy.imaging.projects',
 'platipy.imaging.projects.bronchus',
 'platipy.imaging.projects.cardiac',
 'platipy.imaging.projects.multiatlas',
 'platipy.imaging.projects.nnunet',
 'platipy.imaging.registration',
 'platipy.imaging.tests',
 'platipy.imaging.utils',
 'platipy.imaging.visualisation',
 'platipy.imaging.visualisation.tests']

package_data = \
{'': ['*'],
 'platipy.backend': ['static/css/*',
                     'static/feather/*',
                     'static/feather/icons/*',
                     'static/js/*',
                     'static/pages/*',
                     'templates/*']}

install_requires = \
['SimpleITK>=2.0.2,<3.0.0',
 'click>=8.0.3,<9.0.0',
 'matplotlib>=3.2.2,<4.0.0',
 'pandas>=2.0.3,<3.0.0',
 'pydicom>=2.1.2,<3.0.0',
 'pynetdicom>=2.0.2,<3.0.0',
 'requests>=2.25.1,<3.0.0',
 'rt-utils>=1.1.4,<2.0.0',
 'scikit-image>=0.18.1']

extras_require = \
{'backend': ['Flask>=2.1.0,<3.0.0',
             'Flask-RESTful>=0.3.8',
             'Flask-SQLAlchemy>=2.4.4,<3.0.0',
             'celery>=5.2.3,<6.0.0',
             'redis>=3.5.3,<5.0.0',
             'psutil>=5.8.0,<6.0.0',
             'gunicorn>=20.0.4,<21.0.0',
             'Jinja2>=3.1,<4.0',
             'pymedphys>=0.38.0'],
 'cardiac': ['vtk>=9.1.0,<10.0.0', 'nnunet>=1.7.0,<2.0.0'],
 'nnunet': ['nnunet>=1.7.0,<2.0.0']}

entry_points = \
{'console_scripts': ['platipy = platipy.cli.run:platipy_cli']}

setup_kwargs = {
    'name': 'platipy',
    'version': '0.7.1',
    'description': 'Processing Library and Analysis Toolkit for Medical Imaging in Python',
    'long_description': '# PlatiPy\n\n[![DOI](https://joss.theoj.org/papers/10.21105/joss.05374/status.svg)](https://doi.org/10.21105/joss.05374)\n\n## Processing Library and Analysis Toolkit for Medical Imaging in Python\n\nPlatiPy is a library of **amazing** tools for image processing and analysis - designed specifically\nfor medical imaging!\n\nCheck out the [PlatiPy documentation](https://pyplati.github.io/platipy/) for more info.\n\nThis project was motivated by the need for a simple way to use, visualise, process, and analyse\nmedical images. Many of the tools and algorithms are designed in the context of radiation therapy,\nalthough they are more widely applicable to other fields that use 2D, 3D, or 4D imaging.\n\nPlatiPy is written in Python, and uses SimpleITK, VTK, and standard Python libraries. Jupyter\nnotebooks are provided where possible, mainly for guidance on getting started with using the tools.\nWe welcome feedback and contributions from the community (yes, you!) and you can find more\ninformation about contributing [here](https://pyplati.github.io/platipy/contributing.html).\n\n## What can I do with **platipy**?\n\nA lot! A good place to start is by looking in the\n[examples directory](https://github.com/pyplati/platipy/tree/master/examples).\n\nSome examples of what PlatiPy can do:\n\n- DICOM organising and converting:\n  - Bulk convert from multiple series and studies with a single function\n  - Convert DICOM-RT structure and dose files to NIfTI images\n  - Create DICOM-RT structure files from binary masks e.g. from automatic contouring algorithms\n- Image registration\n  - Register images and transform labels with a few lines of code\n  - Linear transformations: rigid, affine, similarity\n  - Non-linear deformable transformations: demons, b-splines\n  - Multiple metrics for optimisation\n- Atlas-based segmentation\n  - A suite of tools that can be used out-of-the-box\n  - Includes advanced algorithms for\n      [iterative atlas selection](https://doi.org/10.1088/1361-6560/ab652a/) and\n      [vessel splining](https://doi.org/10.1088/1361-6560/abcb1d/)\n- Synthetic deformation field generation\n  - Simulate anatomically realistic shifts, expansions, and bending\n  - Compare DIR results from clinical systems\n- Basic tools for image processing and analysis\n  - Computing label similarity metrics: DSC, mean distance to agreement, Hausdorff distance, and more\n  - Cropping images to a region of interest\n  - Rotate images and generate maximum/mean intensity projections (beams eye view modelling)\n\nA major part of this package is **visualisation**, and some examples are shown below!\n\n#### Visualise some contours\n\n``` python\nfrom platipy.imaging import ImageVisualiser\n\nvis = ImageVisualiser(image)\nvis.add_contour(contours)\nfig = vis.show()\n```\n\n![Figure 1](assets/figure_1.png)\n\n#### Register some images\n\n```python\nfrom platipy.imaging.registration.linear import linear_registration\n\nimage_2_registered, tfm = linear_registration(\nimage_1,\nimage_2\n)\n\nvis = ImageVisualiser(image_1)\nvis.add_comparison_overlay(image_2_registered)\nfig = vis.show()\n```\n\n![Figure 2](assets/figure_2.png)\n\n#### Calculate deformation vector fields\n\n```python\nfrom platipy.imaging.registration.deformable import fast_symmetric_forces_demons_registration\n\nimage_2_deformed, tfm_dir, dvf = fast_symmetric_forces_demons_registration(\nimage_1,\nimage_2_registered\n)\n\nvis = ImageVisualiser(image_2_deformed, axis="z")\nvis.add_vector_overlay(\n    dvf,\n    subsample=12,\n    arrow_scale=1,\n    arrow_width=2,\n    colormap=plt.cm.magma,\n    name="DVF magnitude [mm]",\n    color_function="magnitude"\n)\nfig = vis.show()\n```\n\n![Figure 3](assets/figure_3.png)\n\n## Getting started\n\nThere aren\'t many requirements, just an installed Python interpreter (3.7 or greater). PlatiPy can\nbe installed with **pip**:\n\n```bash\npip install platipy\n```\n\nThe base installation of platipy does not include some large libraries needed for various\ncomponents of platipy. The following extras are available to install to run specific platipy tools:\n\n```bash\npip install platipy[cardiac]\npip install platipy[nnunet]\npip install platipy[backend]\n```\n\n## Authors\n\n- **Phillip Chlap** - [phillip.chlap@unsw.edu.au](phillip.chlap@unsw.edu.au)\n- **Robert Finnegan** - [robert.finnegan@sydney.edu.au](robert.finnegan@sydney.edu.au)\n',
    'author': 'Phillip Chlap & Robert Finnegan',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import os
import tempfile
import time
import uuid
import werkzeug

import logging

import flask_restful
from flask_restful import Api, reqparse
from flask import request, send_from_directory

from functools import wraps

from platipy.backend import app, api
from platipy.dicom.communication import DicomConnector

from .models import db, AlchemyEncoder, APIKey, Dataset, DataObject, DicomLocation
from .tasks import run_task, retrieve_task

logger = logging.getLogger(__name__)


class CustomConfig(object):
    RESTFUL_JSON = {"separators": (", ", ": "), "indent": 2, "cls": AlchemyEncoder}


def authenticate(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not getattr(func, "authenticated", True):
            return func(*args, **kwargs)

        if "API_KEY" in request.headers:
            if APIKey.query.filter_by(key=request.headers["API_KEY"]).first():
                return func(*args, **kwargs)

        flask_restful.abort(401)

    return wrapper


class Resource(flask_restful.Resource):
    method_decorators = [authenticate]  # applies to all inherited resources


class TaskStatus(Resource):
    def get(self, task_id):
        """Get the status of a task given the ID"""
        task = run_task.AsyncResult(task_id)
        if task.state == "PENDING":
            response = {
                "state": task.state,
                "current": 0,
                "total": 1,
                "status": "Pending...",
            }
        elif task.state != "FAILURE":

            if task.info:
                response = {
                    "state": task.state,
                    "current": task.info.get("current", 0),
                    "total": task.info.get("total", 1),
                    "status": task.info.get("status", ""),
                }
                if "result" in task.info:
                    response["result"] = task.info["result"]
                if "series" in task.info:
                    response["series"] = task.info["series"]
            else:
                response = {"state": task.state}

        else:
            # something went wrong in the background job
            response = {
                "state": task.state,
                "current": 1,
                "total": 1,
                "status": str(task.info),  # this is the exception raised
            }
        return response


class DicomLocationEndpoint(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        "name",
        required=True,
        help="Name to identify this Dicom location",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "host",
        required=True,
        help="Dicom location host name or IP address",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "port",
        type=int,
        required=True,
        help="The port of the Dicom location",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "ae_title", help="AE Title of the Dicom location", location=["args", "headers", "values"]
    )

    def get(self):

        key = request.headers["API_KEY"]

        dl = DicomLocation.query.filter_by(owner_key=key).all()

        return dl

    def post(self):

        args = self.parser.parse_args()

        key = request.headers["API_KEY"]
        dl = DicomLocation(
            owner_key=key,
            name=args["name"],
            host=args["host"],
            port=args["port"],
            ae_title=args["ae_title"],
        )

        db.session.add(dl)
        db.session.commit()

        return dl


class DataObjectsEndpoint(Resource):
    def get(self):

        key = request.headers["API_KEY"]

        do = DataObject.query.filter(DataObject.dataset.has(owner_key=key)).all()

        return do


class DataObjectEndpoint(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        "dataset",
        required=True,
        help="Dataset ID to add Data Object to",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "type",
        choices=("DICOM", "FILE"),
        required=True,
        help="DICOM for Dicom objects to be fetched from the Dataset Dicom Location. FILE for file sent with request.",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "dicom_retrieve",
        choices=("MOVE", "GET", "SEND"),
        help="Used for DICOM type. The Dicom objects will be retrieved using this method.",
        location=["args", "headers", "values"],
    )
    parser.add_argument("seriesUID", location=["args", "headers", "values"])
    parser.add_argument("meta_data", location=["args", "headers", "values"])
    parser.add_argument("file_name", location=["args", "headers", "values"])
    parser.add_argument("file_data", type=werkzeug.datastructures.FileStorage, location="files")
    parser.add_argument(
        "parent",
        help="Data Object ID to which this data object should be linked",
        location=["args", "headers", "values"],
    )

    def get(self, dataobject_id):

        key = request.headers["API_KEY"]

        do = DataObject.query.filter(
            DataObject.dataset.has(owner_key=key), DataObject.id == dataobject_id
        ).first()

        if do:
            return do

        return {"Error": "Data Object not found"}, 404

    def post(self):

        key = request.headers["API_KEY"]

        args = self.parser.parse_args()
        dataset_id = args["dataset"]

        # Get the dataset to which this data object should be added
        ds = Dataset.query.filter_by(owner_key=key, id=dataset_id).first()
        if not ds:
            return {"Error": "Dataset not found"}, 404

        # Get the parent dataset if one was given
        parent = None
        if args["parent"]:
            parent = DataObject.query.filter_by(dataset_id=ds.id, id=args["parent"]).first()

            if not parent:
                return {"Error": "Parent Data Object not found"}, 404

        meta_data = None
        if args["meta_data"]:
            meta_data = json.loads(args["meta_data"])

        # Create the DataObject
        do = DataObject(
            dataset=ds,
            is_input=True,
            type=args["type"],
            series_instance_uid=args["seriesUID"],
            meta_data=meta_data,
            parent=parent,
        )
        db.session.add(do)
        db.session.commit()

        if args["type"] == "DICOM":

            dicom_fetch = args["dicom_retrieve"]
            if not dicom_fetch:
                return (
                    {
                        "message": {
                            "dicom_retrieve": "Set GET, MOVE or SEND to be able to retrieve Dicom objects."
                        }
                    },
                    400,
                )

            if not args["seriesUID"]:
                return (
                    {
                        "message": {
                            "seriesUID": "SeriesUID is required to be able to retrieve DICOM objects"
                        }
                    },
                    400,
                )

            if dicom_fetch == "MOVE":

                if not ds.from_dicom_location:
                    return (
                        {
                            "message": {
                                "from_dicom_location": "Dataset From Dicom Location not set, so unable to MOVE DICOM objects"
                            }
                        },
                        400,
                    )

                # Fetch Dicom data using MOVE
                # Check whether or not we are listening for for Dicom MOVE
                listening_connector = DicomConnector(
                    host="127.0.0.1",
                    port=app.dicom_listener_port,
                    ae_title=app.dicom_listener_aetitle,
                )

                if not listening_connector.verify():

                    # Verify Dicom Location is listening
                    timeout_seconds = 20
                    time_waited = 0

                    # We are not listening, wait for 20 seconds and abort if still not listening
                    while not listening_connector.verify():
                        logger.debug(
                            "Not listening for MOVE, sleeping for 1 second and will try again"
                        )
                        time.sleep(1)
                        time_waited += 1

                        if time_waited >= timeout_seconds:
                            msg = "Listener for MOVE timeout on port: {0}".format(
                                ds.from_dicom_location.move_port
                            )
                            logger.error(msg)
                            return {"message": {"from_dicom_location": msg}}, 400

                    logger.info("Listening for MOVE OK")

                # Trigger MOVE
                logger.info(
                    "Triggering MOVE at %s for series UID: %s",
                    app.dicom_listener_aetitle,
                    do.series_instance_uid,
                )
                dicom_connector = DicomConnector(
                    host=ds.from_dicom_location.host,
                    port=ds.from_dicom_location.port,
                    ae_title=ds.from_dicom_location.ae_title,
                )

                dicom_verify = dicom_connector.verify()

                if dicom_verify:
                    dicom_connector.move_series(
                        do.series_instance_uid, move_aet=app.dicom_listener_aetitle
                    )
                else:
                    msg = "Unable to connect to Dicom Location: {0} {1} {2}".format(
                        ds.from_dicom_location.host,
                        ds.from_dicom_location.port,
                        ds.from_dicom_location.ae_title,
                    )
                    logger.error(msg)
                    return {"message": {"from_dicom_location": msg}}, 400

            elif dicom_fetch == "GET":

                if not ds.from_dicom_location:
                    return (
                        {
                            "message": {
                                "from_dicom_location": "Dataset From Dicom Location not set, so unable to GET DICOM objects"
                            }
                        },
                        400,
                    )

                # Fetch Dicom data using GET
                task = retrieve_task.apply_async([do.id])

            # If dicom_fetch is SEND we don't do anything here, just wait for the client
            # to send to our Dicom Listener.

        elif args["type"] == "FILE":

            if not args["file_name"]:
                return {"message": {"file_name": "Provide the file name"}}, 400

            if not args["file_data"]:
                return {"message": {"file_data": "Provide the file data"}}, 400

            # Save the file
            file_path = os.path.join(tempfile.mkdtemp(), args["file_name"])
            args["file_data"].save(file_path)
            do.is_fetched = True
            do.path = file_path

            db.session.add(do)
            db.session.commit()

        return do

    def delete(self, dataobject_id):

        key = request.headers["API_KEY"]

        do = DataObject.query.filter(
            DataObject.dataset.has(owner_key=key), DataObject.id == dataobject_id
        ).first()

        if do:
            db.session.delete(do)
            db.session.commit()

        return 200


class DataObjectDownloadEndpoint(Resource):

    parser = reqparse.RequestParser()

    def get(self, dataobject_id):
        """Returns the file for the given dataobject"""

        key = request.headers["API_KEY"]

        do = DataObject.query.filter(
            DataObject.dataset.has(owner_key=key), DataObject.id == dataobject_id
        ).first()

        if do:

            if not do.type == "FILE":
                return {"Error": "Can only download Data Objects of type FILE"}, 400

            f = do.path

            if not os.path.exists(f):
                return {"Error": "File could not be found, perhaps it has expired"}, 404

            logger.info("Downloading file: %s", f)
            return send_from_directory(os.path.dirname(f), os.path.basename(f), as_attachment=True)

        return {"Error": "Data Object not found"}, 404


class DatasetsEndpoint(Resource):
    def get(self):

        key = request.headers["API_KEY"]

        ds = Dataset.query.filter_by(owner_key=key).all()

        return ds


class DatasetEndpoint(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        "from_dicom_location",
        help="ID of DicomLocation from which to retrieve DICOM data",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "to_dicom_location",
        help="ID of DicomLocation the send output data to",
        location=["args", "headers", "values"],
    )
    parser.add_argument("timeout", type=int, default=24, location=["args", "headers", "values"])

    def get(self, dataset_id):

        key = request.headers["API_KEY"]

        ds = Dataset.query.filter_by(owner_key=key, id=dataset_id).first()

        if ds:
            return ds

        return {"Error": "Dataset not found"}, 404

    def post(self):

        args = self.parser.parse_args()

        key = request.headers["API_KEY"]
        expiry = datetime.datetime.now() + datetime.timedelta(hours=args["timeout"])
        ds = Dataset(owner_key=key, timeout=expiry)

        if args["from_dicom_location"]:
            from_dicom_location = DicomLocation.query.filter_by(
                owner_key=key, id=args["from_dicom_location"]
            ).first()
            if not from_dicom_location:
                return {"Error": "From Dicom Location not found"}, 404
            ds.from_dicom_location = from_dicom_location

        if args["to_dicom_location"]:
            to_dicom_location = DicomLocation.query.filter_by(
                owner_key=key, id=args["to_dicom_location"]
            ).first()
            if not to_dicom_location:
                return {"Error": "To Dicom Location not found"}, 404
            ds.to_dicom_location = to_dicom_location

        db.session.add(ds)
        db.session.commit()

        return ds


class DatasetReadyEndpoint(Resource):

    parser = reqparse.RequestParser()

    def get(self, dataset_id):

        key = request.headers["API_KEY"]

        ds = Dataset.query.filter_by(owner_key=key, id=dataset_id).first()

        if ds:
            if len(ds.input_data_objects) == 0:
                return {"ready": False}, 200

            for d in ds.input_data_objects:
                if not d.is_fetched:
                    return {"ready": False}, 200

            return {"ready": True}, 200

        return {"Error": "Dataset not found"}, 404


class AlgorithmEndpoint(Resource):
    def get(self):

        result = []
        for a in app.algorithms:
            result.append({"name": a, "default_settings": app.algorithms[a].default_settings})
        return result


class TriggerEndpoint(Resource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        "algorithm",
        required=True,
        help="The name of the algorithm to trigger",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "dataset",
        required=True,
        help="The ID of the dataset to pass to the algorithm",
        location=["args", "headers", "values"],
    )
    parser.add_argument(
        "config",
        help="JSON configuration for algorithm. Default configuration will be used if not set.",
        location=["args", "headers", "values"],
    )

    def post(self):

        args = self.parser.parse_args()

        if not args["algorithm"] in app.algorithms:
            return (
                {"Error": "No algorithm found with name: {0}".format(args["algorithm"])},
                404,
            )

        algorithm = app.algorithms[args["algorithm"]]

        config = algorithm.default_settings
        if args["config"]:
            try:
                config = json.loads(args["config"])
            except json.decoder.JSONDecodeError:
                return {"Error": "Could not parse JSON config given"}, 500

        key = request.headers["API_KEY"]
        ds = Dataset.query.filter_by(owner_key=key, id=args["dataset"]).first()
        if not ds:
            return {"Error": "Dataset not found"}, 404

        # Ensure that all input data objects in the dataset have been fetched
        if len(ds.input_data_objects) == 0:
            return {"Error": "Dataset does not contain any data objects"}, 400

        for d in ds.input_data_objects:
            if not d.is_fetched:
                return (
                    {
                        "Error": "Dataset contains Data Objects that have not been fetched yet. Wait for them to be fetched or DICOM Send them if necessary."
                    },
                    400,
                )

        # Start the algorithm task
        task = run_task.apply_async([algorithm.name, config, ds.id])

        # Return JSON data detailing where to poll for updates on the task
        return {"poll": api.url_for(TaskStatus, task_id=task.id)}


api.add_resource(TaskStatus, "/api/status/<string:task_id>")
api.add_resource(TriggerEndpoint, "/api/trigger")

api.add_resource(DatasetsEndpoint, "/api/datasets")
api.add_resource(DatasetEndpoint, "/api/dataset", "/api/dataset/<string:dataset_id>")
api.add_resource(DatasetReadyEndpoint, "/api/dataset/ready/<string:dataset_id>")

api.add_resource(DataObjectsEndpoint, "/api/dataobjects")
api.add_resource(DataObjectEndpoint, "/api/dataobject", "/api/dataobject/<string:dataobject_id>")
api.add_resource(DataObjectDownloadEndpoint, "/api/dataobject/download/<string:dataobject_id>")

api.add_resource(AlgorithmEndpoint, "/api/algorithm")

api.add_resource(
    DicomLocationEndpoint,
    "/api/dicomlocation",
    "/api/dicomlocation/<string:dicom_location_id>",
)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk


def insert_sphere(arr, sp_radius=4, sp_centre=(0, 0, 0)):
    """Insert a sphere into the given array

    Args:
        arr (np.array): Array in which to insert sphere
        sp_radius (int, optional): The radius of the sphere. Defaults to 4.
        sp_centre (tuple, optional): The position at which the sphere should be inserted. Defaults
                                     to (0, 0, 0).

    Returns:
        np.array: An array with the sphere inserted
    """

    arr_copy = arr[:]

    x, y, z = np.indices(arr.shape)

    if not hasattr(sp_radius, "__iter__"):
        sp_radius = [sp_radius] * 3

    sp_radius_x, sp_radius_y, sp_radius_z = sp_radius

    arr_copy[
        ((x - sp_centre[0]) / sp_radius_x) ** 2.0
        + ((y - sp_centre[1]) / sp_radius_y) ** 2.0
        + ((z - sp_centre[2]) / sp_radius_z) ** 2.0
        <= 1
    ] = 1

    return arr_copy


def insert_cylinder(arr, cyl_radius=4, cyl_height=2, cyl_centre=(0, 0, 0)):
    """
    Insert a cylinder into the given array.
    The cylinder vertical extent is +/- 0.5 * height

    Args:
        arr (np.ndarray): The array into which the cylinder is inserted
        cyl_radius (int, optional): Cylinder radius. Defaults to 4.
        cyl_height (int, optional): Cylinder height. Defaults to 2.
        cyl_centre (tuple, optional): Cylinder centre. Defaults to (0, 0, 0).

    Returns:
        np.ndarray: The original array with a cylinder (value 1)
    """
    arr_copy = arr[:]

    x, y, z = np.indices(arr.shape)

    if not hasattr(cyl_radius, "__iter__"):
        cyl_radius = [cyl_radius] * 2

    condition_radial = (
        ((z - cyl_centre[0]) / cyl_radius[0]) ** 2 + ((y - cyl_centre[1]) / cyl_radius[1]) ** 2
    ) <= 1
    condition_height = np.abs((x - cyl_centre[2]) / (0.5 * cyl_height)) <= 1

    arr_copy[condition_radial & condition_height] = 1

    return arr_copy


def insert_sphere_image(image, sp_radius, sp_centre):
    """Insert a sphere into an image

    Args:
        image (sitk.Image): Image in which to insert sphere
        sp_radius (int | list, optional): The radius of the sphere.
            Can also be defined as a vector. Defaults to 4.
        sp_centre (tuple, optional): The position at which the sphere should be inserted. Defaults
            to (0, 0, 0).

    Returns:
        np.array: An array with the sphere inserted
    """

    if not hasattr(sp_radius, "__iter__"):
        sp_radius = [sp_radius] * 3

    sp_radius_image = [i / j for i, j in zip(sp_radius, image.GetSpacing()[::-1])]

    arr = sitk.GetArrayFromImage(image)

    arr = insert_sphere(arr, sp_radius_image, sp_centre)

    image_sphere = sitk.GetImageFromArray(arr)
    image_sphere.CopyInformation(image)

    return image_sphere


def insert_cylinder_image(image, cyl_radius=(5, 5), cyl_height=10, cyl_centre=(0, 0, 0)):
    """Insert a cylinder into an image

    Args:
        image (SimpleITK.Image):
        cyl_radius (tuple, optional): Cylinder radius, can be defined as a single value
            or a tuple (will generate an ellipsoid). Defaults to (5,5).
        cyl_height (int, optional): Cylinder height. Defaults to 10.
        cyl_centre (tuple, optional): Cylinder centre. Defaults to (0,0,0).

    Returns:
        SimpleITK.Image: Image with cylinder inserted
    """
    if not hasattr(cyl_radius, "__iter__"):
        cyl_radius = [cyl_radius] * 2

    cyl_radius_image = [i / j for i, j in zip(cyl_radius, image.GetSpacing()[1::-1])]
    cyl_height_image = cyl_height / image.GetSpacing()[2]

    arr = sitk.GetArrayFromImage(image)

    arr = insert_cylinder(arr, cyl_radius_image, cyl_height_image, cyl_centre)

    image_cylinder = sitk.GetImageFromArray(arr)
    image_cylinder.CopyInformation(image)

    return image_cylinder



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
import tempfile
import pydicom

from pydicom.dataset import Dataset, FileDataset

from pydicom.uid import ImplicitVRLittleEndian
from pynetdicom import (
    AE,
    evt,
    VerificationPresentationContexts,
    StoragePresentationContexts,
    QueryRetrievePresentationContexts,
    PYNETDICOM_IMPLEMENTATION_UID,
    PYNETDICOM_IMPLEMENTATION_VERSION,
)
from pynetdicom.sop_class import (
    Verification,
    PatientRootQueryRetrieveInformationModelFind,
    PatientRootQueryRetrieveInformationModelGet,
    PatientRootQueryRetrieveInformationModelMove,
)
from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

logger = logging.getLogger(__name__)


class DicomConnector:
    def __init__(self, host="127.0.0.1", port=0, ae_title="", output_directory=tempfile.mkdtemp()):
        self.host = host
        self.port = port
        self.ae_title = ae_title if ae_title else ""

        logger.debug(
            "DicomConnector with host: %s port: %d AETitle: %s",
            self.host,
            self.port,
            self.ae_title,
        )

        self.output_directory = output_directory
        self.current_dir = None
        self.recieved_callback = None

    def verify(self):
        # Verify Connection

        ae = AE()
        ae.requested_contexts = VerificationPresentationContexts

        # Associate with a peer DICOM AE
        if len(self.ae_title) > 0:
            assoc = ae.associate(self.host, self.port, ae_title=self.ae_title)
        else:
            assoc = ae.associate(self.host, self.port)

        result = None

        if assoc.is_established:
            status = assoc.send_c_echo()

            if status:
                result = status.Status

            # Release the association
            assoc.release()

        return not result is None

    def do_find(self, dataset, query_model=PatientRootQueryRetrieveInformationModelFind):

        ae = AE()

        ae.requested_contexts = QueryRetrievePresentationContexts

        if len(self.ae_title) > 0:
            assoc = ae.associate(self.host, self.port, ae_title=self.ae_title)
        else:
            assoc = ae.associate(self.host, self.port)

        results = []
        if assoc.is_established:
            logger.info("Association accepted by the peer")

            responses = assoc.send_c_find(dataset, query_model=query_model)

            for _, ds in responses:
                results.append(ds)

            # Release the association
            assoc.release()

        logger.info("Got %d results", len(results))

        return results

    def get_studies_for_patient(self, patient_id):

        dataset = Dataset()
        dataset.StudyInstanceUID = ""
        dataset.StudyDescription = ""
        dataset.PatientID = patient_id
        dataset.PatientName = ""
        dataset.QueryRetrieveLevel = "STUDY"

        return self.do_find(dataset)

    def get_series_for_study(self, study_instance_uid, modality):

        dataset = Dataset()
        dataset.StudyInstanceUID = study_instance_uid
        dataset.Modality = modality
        dataset.SeriesInstanceUID = ""
        dataset.SeriesDescription = ""
        dataset.QueryRetrieveLevel = "SERIES"

        return self.do_find(dataset)

    def move_series(
        self,
        seriesInstanceUID,
        move_aet="PYNETDICOM",
        query_model=PatientRootQueryRetrieveInformationModelMove,
    ):

        ae = AE()
        ae.requested_contexts = QueryRetrievePresentationContexts

        if len(self.ae_title) > 0:
            assoc = ae.associate(self.host, self.port, ae_title=self.ae_title)
        else:
            assoc = ae.associate(self.host, self.port)

        if assoc.is_established:
            logger.info("Association accepted by the peer")

            dataset = Dataset()
            dataset.SeriesInstanceUID = seriesInstanceUID
            dataset.QueryRetrieveLevel = "SERIES"

            responses = assoc.send_c_move(dataset, move_aet, query_model=query_model)

            for (_, _) in responses:
                pass

            # Release the association
            assoc.release()

        logger.info("Finished")

    def download_series(
        self,
        series_instance_uid,
        recieved_callback=None,
        query_model=PatientRootQueryRetrieveInformationModelGet,
    ):

        self.recieved_callback = recieved_callback

        ae = AE()

        # Specify which SOP Classes are supported as an SCU
        for context in QueryRetrievePresentationContexts:
            ae.add_requested_context(context.abstract_syntax, ImplicitVRLittleEndian)
        for context in StoragePresentationContexts[:115]:
            ae.add_requested_context(context.abstract_syntax, ImplicitVRLittleEndian)

        # Add SCP/SCU Role Selection Negotiation to the extended negotiation
        # We want to act as a Storage SCP
        ext_neg = []
        for context in StoragePresentationContexts:
            role = SCP_SCU_RoleSelectionNegotiation()
            role.sop_class_uid = context.abstract_syntax
            role.scp_role = True
            role.scu_role = False
            ext_neg.append(role)

        handlers = [(evt.EVT_C_STORE, self.on_c_store)]

        if len(self.ae_title) > 0:
            assoc = ae.associate(
                self.host,
                self.port,
                ae_title=self.ae_title,
                ext_neg=ext_neg,
                evt_handlers=handlers,
            )
        else:
            assoc = ae.associate(self.host, self.port, ext_neg=ext_neg, evt_handlers=handlers)

        if assoc.is_established:
            logger.info("Association accepted by the peer")

            dataset = Dataset()
            dataset.SeriesInstanceUID = series_instance_uid
            dataset.QueryRetrieveLevel = "SERIES"

            responses = assoc.send_c_get(dataset, query_model=query_model)

            for (_, _) in responses:
                pass

            # Release the association
            assoc.release()

        logger.info("Finished")

        return self.current_dir

    def on_c_store(self, event):

        dataset = event.dataset

        mode_prefix = "UN"
        mode_prefixes = {
            "CT Image Storage": "CT",
            "Enhanced CT Image Storage": "CTE",
            "MR Image Storage": "MR",
            "Enhanced MR Image Storage": "MRE",
            "Positron Emission Tomography Image Storage": "PT",
            "Enhanced PET Image Storage": "PTE",
            "RT Image Storage": "RI",
            "RT Dose Storage": "RD",
            "RT Plan Storage": "RP",
            "RT Structure Set Storage": "RS",
            "Computed Radiography Image Storage": "CR",
            "Ultrasound Image Storage": "US",
            "Enhanced Ultrasound Image Storage": "USE",
            "X-Ray Angiographic Image Storage": "XA",
            "Enhanced XA Image Storage": "XAE",
            "Nuclear Medicine Image Storage": "NM",
            "Secondary Capture Image Storage": "SC",
        }

        try:
            mode_prefix = mode_prefixes[dataset.SOPClassUID.name]
        except KeyError:
            mode_prefix = "UN"

        suid = dataset.SeriesInstanceUID
        series_dir = suid
        if self.output_directory is not None:
            series_dir = os.path.join(self.output_directory, suid)

        filename = "{0!s}.{1!s}".format(mode_prefix, dataset.SOPInstanceUID)

        if not os.path.exists(series_dir):
            os.mkdir(series_dir)

        filename = os.path.join(series_dir, filename)

        if os.path.exists(filename):
            logger.debug("DICOM file already exists, overwriting")

        context = event.context
        meta = Dataset()
        meta.MediaStorageSOPClassUID = dataset.SOPClassUID
        meta.MediaStorageSOPInstanceUID = dataset.SOPInstanceUID
        meta.ImplementationClassUID = PYNETDICOM_IMPLEMENTATION_UID
        meta.TransferSyntaxUID = context.transfer_syntax

        # The following is not mandatory, set for convenience
        meta.ImplementationVersionName = PYNETDICOM_IMPLEMENTATION_VERSION
        file_ds = FileDataset(filename, {}, file_meta=meta, preamble=b"\0" * 128)
        file_ds.update(dataset)
        file_ds.is_little_endian = context.transfer_syntax.is_little_endian
        file_ds.is_implicit_VR = context.transfer_syntax.is_implicit_VR

        status_ds = Dataset()
        status_ds.Status = 0x0000

        try:
            # We use `write_like_original=False` to ensure that a compliant
            #   File Meta Information Header is written
            file_ds.save_as(filename, write_like_original=False)
            status_ds.Status = 0x0000  # Success
        except IOError:
            logger.warning("Could not write file to specified directory:")
            logger.warning("    %s", os.path.dirname(filename))
            logger.warning("Directory may not exist or you may not have write " "permission")
            # Failed - Out of Resources - IOError
            status_ds.Status = 0xA700
        except Exception as exception:
            logger.warning("Could not write file to specified directory:")
            logger.warning("    %s", os.path.dirname(filename))
            logger.warning(exception)
            # Failed - Out of Resources - Miscellaneous error
            status_ds.Status = 0xA701

        self.current_dir = series_dir

        return status_ds

    def send_dcm(self, dcm_file):
        """
        send_dcm accepts the location of a single dcm file to send over the connector or a list
        of dcm files to send in one association.
        """

        dcm_files = dcm_file
        if isinstance(dcm_file, str):
            dcm_files = [dcm_file]

        transfer_syntax = [ImplicitVRLittleEndian]

        ae = AE()

        for context in StoragePresentationContexts:
            ae.add_requested_context(context.abstract_syntax, transfer_syntax)

        if len(self.ae_title) > 0:
            assoc = ae.associate(self.host, self.port, ae_title=self.ae_title)
        else:
            assoc = ae.associate(self.host, self.port)

        status = ""
        if assoc.is_established:
            logger.debug("Sending file: %s", dcm_file)

            for dcm_file in dcm_files:
                dataset = pydicom.read_file(dcm_file)
                status = assoc.send_c_store(dataset)

            # Release the association
            assoc.release()

        return status

    def on_c_echo(self, event):
        """Respond to a C-ECHO service request.

        Parameters
        ----------
        context : namedtuple
            The presentation context that the verification request was sent under.
        info : dict
            Information about the association and verification request.

        Returns
        -------
        status : int or pydicom.dataset.Dataset
            The status returned to the peer AE in the C-ECHO response. Must be
            a valid C-ECHO status value for the applicable Service Class as
            either an ``int`` or a ``Dataset`` object containing (at a
            minimum) a (0000,0900) *Status* element.
        """
        logger.debug("C-ECHO!")
        return 0x0000

    def on_association_accepted(self, event):
        self.current_dir = None

    def on_association_released(self, event):

        if self.recieved_callback and self.current_dir:
            self.recieved_callback(self.current_dir)

    def listen(self, recieved_callback, ae_title="PYNETDICOM"):

        self.recieved_callback = recieved_callback

        # Initialise the Application Entity and specify the listen port
        ae = AE(ae_title=ae_title)

        # Add the supported presentation context
        ae.add_supported_context(Verification)
        for context in StoragePresentationContexts:
            ae.add_supported_context(context.abstract_syntax)

        handlers = [
            (evt.EVT_C_MOVE, self.on_c_store),
            (evt.EVT_C_STORE, self.on_c_store),
            (evt.EVT_C_ECHO, self.on_c_echo),
            (evt.EVT_ACCEPTED, self.on_association_accepted),
            (evt.EVT_RELEASED, self.on_association_released),
        ]

        # Start listening for incoming association requests
        ae.start_server(("", self.port), evt_handlers=handlers)
